import sys
from mock import patch
import tempfile
from click.testing import CliRunner

import wily.__main__ as main

_path = "src\\test.py" if sys.platform == "win32" else "src/test.py"


PATCHED_ENV = {
    "BROWSER": "echo %s",
    "LC_ALL": "C.UTF-8",
    "LANG": "C.UTF-8",
    "HOME": tempfile.gettempdir(),
}


def test_graph_no_cache(tmpdir, cache_path):
    runner = CliRunner()
    with patch.dict("os.environ", values=PATCHED_ENV, clear=True):
        result = runner.invoke(
            main.cli,
            ["--path", tmpdir, "--cache", cache_path, "graph", _path, "raw.loc"],
        )
    assert result.exit_code == 1, result.stdout


def test_graph(builddir):
    """ Test the graph feature """
    runner = CliRunner()
    with patch.dict("os.environ", values=PATCHED_ENV, clear=True):
        result = runner.invoke(
            main.cli, ["--path", builddir, "graph", _path, "raw.loc"]
        )
    assert result.exit_code == 0, result.stdout


def test_graph_all(builddir):
    """ Test the graph feature """
    runner = CliRunner()
    with patch.dict("os.environ", values=PATCHED_ENV, clear=True):
        result = runner.invoke(
            main.cli, ["--path", builddir, "graph", _path, "raw.loc", "--all"]
        )
    assert result.exit_code == 0, result.stdout


def test_graph_all(builddir):
    """ Test the graph feature with shorthand metric"""
    runner = CliRunner()
    with patch.dict("os.environ", values=PATCHED_ENV, clear=True):
        result = runner.invoke(
            main.cli, ["--path", builddir, "graph", _path, "loc", "--all"]
        )
    assert result.exit_code == 0, result.stdout


def test_graph_changes(builddir):
    """ Test the graph feature """
    runner = CliRunner()
    with patch.dict("os.environ", values=PATCHED_ENV, clear=True):
        result = runner.invoke(
            main.cli, ["--path", builddir, "graph", _path, "raw.loc", "--changes"]
        )
    assert result.exit_code == 0, result.stdout


def test_graph_custom_x(builddir):
    """ Test the graph feature """
    runner = CliRunner()
    with patch.dict("os.environ", values=PATCHED_ENV, clear=True):
        result = runner.invoke(
            main.cli, ["--path", builddir, "graph", _path, "raw.loc", "-x", "raw.sloc"]
        )
    assert result.exit_code == 0, result.stdout


def test_graph_path(builddir):
    """ Test the graph feature """
    runner = CliRunner()
    with patch.dict("os.environ", values=PATCHED_ENV, clear=True):
        result = runner.invoke(
            main.cli, ["--path", builddir, "graph", "src/", "raw.loc"]
        )
    assert result.exit_code == 0, result.stdout


def test_graph_multiple(builddir):
    """ Test the graph feature with multiple metrics """
    runner = CliRunner()
    with patch.dict("os.environ", values=PATCHED_ENV, clear=True):
        result = runner.invoke(
            main.cli, ["--path", builddir, "graph", _path, "raw.loc", "raw.comments"]
        )
    assert result.exit_code == 0, result.stdout


def test_graph_multiple_custom_x(builddir):
    """ Test the graph feature with multiple metrics """
    runner = CliRunner()
    with patch.dict("os.environ", values=PATCHED_ENV, clear=True):
        result = runner.invoke(
            main.cli,
            [
                "--path",
                builddir,
                "graph",
                _path,
                "raw.loc",
                "raw.comments",
                "-x",
                "raw.sloc",
            ],
        )
    assert result.exit_code == 0, result.stdout


def test_graph_multiple_path(builddir):
    """ Test the graph feature with multiple metrics """
    runner = CliRunner()
    with patch.dict("os.environ", values=PATCHED_ENV, clear=True):
        result = runner.invoke(
            main.cli, ["--path", builddir, "graph", "src/", "raw.loc", "raw.comments"]
        )
    assert result.exit_code == 0, result.stdout


def test_graph_output(builddir):
    """ Test the graph feature with target output file """
    runner = CliRunner()
    with patch.dict("os.environ", values=PATCHED_ENV, clear=True):
        result = runner.invoke(
            main.cli,
            [
                "--debug",
                "--path",
                builddir,
                "graph",
                _path,
                "raw.loc",
                "-o",
                "test.html",
            ],
        )

    assert result.exit_code == 0, result.stdout


def test_graph_output_granular(builddir):
    """ Test the graph feature with target output file """
    runner = CliRunner()
    with patch.dict("os.environ", values=PATCHED_ENV, clear=True):
        result = runner.invoke(
            main.cli,
            [
                "--debug",
                "--path",
                builddir,
                "graph",
                "src/test.py:function1",
                "cyclomatic.complexity",
                "-o",
                "test_granular.html",
            ],
        )
    assert result.exit_code == 0, result.stdout



import pytest

from click.testing import CliRunner
from mock import patch
from pathlib import Path

import wily.__main__ as main

from wily.helper.custom_enums import ReportFormat


def test_init():
    with patch.object(main, "cli", return_value=None) as cli:
        with patch.object(main, "__name__", "__main__"):
            __import__("wily.__main__")
            assert cli.called_once


def test_help():
    """
    Test that CLI when called with help options
    """
    with patch("wily.logger") as logger:
        runner = CliRunner()
        result = runner.invoke(main.cli, ["--help", "--debug"])
        assert result.exit_code == 0


def test_setup():
    """
    Test that CLI when called with help options
    """
    with patch("wily.__main__.handle_no_cache", return_value=True) as handle_no_cache:
        runner = CliRunner()
        result = runner.invoke(main.cli, ["setup"])
        assert result.exit_code == 0
        assert handle_no_cache.called_once


def test_handle_no_cache_no():
    """
    Test that setup cancels when "n" typed
    """
    with patch("wily.__main__.input", return_value="n") as mock_input:
        with pytest.raises(SystemExit):
            main.handle_no_cache(None)
            assert mock_input.called_once


def test_handle_no_cache():
    """
    Test that setup works
    """
    with patch("wily.__main__.build", return_value="n") as build_command:
        with patch("wily.__main__.input", side_effect=["y", "11", "."]) as mock_input:
            runner = CliRunner()
            runner.invoke(main.cli, ["setup"])
            assert mock_input.called
            assert build_command.called_once
            assert build_command.called_with("1")


def test_build():
    """
    Test that build calls the build command
    """
    with patch("wily.commands.build.build") as build:
        runner = CliRunner()
        result = runner.invoke(main.cli, ["build", "wily"])
        assert result.exit_code == 0
        assert build.called_once


def test_build_with_opts():
    """
    Test that build calls the build command
    """
    with patch("wily.commands.build.build") as build:
        runner = CliRunner()
        result = runner.invoke(
            main.cli, ["build", "wily", "-n 1", "-o raw,maintainability"]
        )
        assert result.exit_code == 0
        assert build.called_once
        assert build.call_args[1]["config"].max_revisions == 1
        assert build.call_args[1]["config"].operators == ["raw", "maintainability"]


def test_index():
    """
    Test that index calls the index command
    """
    with patch("wily.__main__.exists", return_value=True) as check_cache:
        with patch("wily.commands.index.index") as index:
            runner = CliRunner()
            result = runner.invoke(main.cli, ["index"])
            assert result.exit_code == 0
            assert index.called_once
            assert check_cache.called_once


def test_index_with_opts():
    """
    Test that index calls the index command with options
    """
    with patch("wily.__main__.exists", return_value=True) as check_cache:
        with patch("wily.commands.index.index") as index:
            runner = CliRunner()
            result = runner.invoke(main.cli, ["index", "--message"])
            assert result.exit_code == 0
            assert index.called_once
            assert check_cache.called_once
            assert index.call_args[1]["include_message"]


def test_index_with_no_message():
    """
    Test that index calls the index command with options
    """
    with patch("wily.__main__.exists", return_value=True) as check_cache:
        with patch("wily.commands.index.index") as index:
            runner = CliRunner()
            result = runner.invoke(main.cli, ["index", "--no-message"])
            assert result.exit_code == 0
            assert index.called_once
            assert check_cache.called_once
            assert not index.call_args[1]["include_message"]


def test_report():
    """
    Test that report calls the report command
    """
    with patch(
        "wily.__main__.get_default_metrics",
        return_value=["maintainability.mi", "raw.loc"],
    ) as gdf:
        with patch("wily.__main__.exists", return_value=True) as check_cache:
            with patch("wily.commands.report.report") as report:
                runner = CliRunner()
                result = runner.invoke(main.cli, ["report", "foo.py"])
                assert result.exit_code == 0, result.stdout
                assert report.called_once
                assert check_cache.called_once
                assert report.call_args[1]["path"] == "foo.py"
                assert report.call_args[1]["format"] == ReportFormat.CONSOLE
                assert "maintainability.mi" in report.call_args[1]["metrics"]
                assert gdf.called_once


def test_report_with_opts():
    """
    Test that report calls the report command with options
    """
    with patch("wily.__main__.exists", return_value=True) as check_cache:
        with patch("wily.commands.report.report") as report:
            runner = CliRunner()
            result = runner.invoke(
                main.cli, ["report", "foo.py", "example_metric", "-n 101", "--message"]
            )
            assert result.exit_code == 0, result.stdout
            assert report.called_once
            assert check_cache.called_once
            assert report.call_args[1]["path"] == "foo.py"
            assert report.call_args[1]["metrics"] == ("example_metric",)
            assert report.call_args[1]["include_message"]
            assert report.call_args[1]["n"] == 101
            assert report.call_args[1]["format"] == ReportFormat.CONSOLE


def test_report_html_format():
    """
    Test that report calls the report command with HTML as format
    """
    with patch("wily.__main__.exists", return_value=True) as check_cache:
        with patch("wily.commands.report.report") as report:
            runner = CliRunner()
            result = runner.invoke(
                main.cli,
                [
                    "report",
                    "foo.py",
                    "example_metric",
                    "-n 101",
                    "--message",
                    "--format=HTML",
                ],
            )

            assert result.exit_code == 0, result.stdout
            assert report.called_once
            assert check_cache.called_once
            assert report.call_args[1]["path"] == "foo.py"
            assert report.call_args[1]["metrics"] == ("example_metric",)
            assert report.call_args[1]["include_message"]
            assert report.call_args[1]["n"] == 101
            assert report.call_args[1]["format"] == ReportFormat.HTML


def test_report_console_format():
    """
    Test that report calls the report command with console as format
    """
    with patch("wily.__main__.exists", return_value=True) as check_cache:
        with patch("wily.commands.report.report") as report:
            runner = CliRunner()
            result = runner.invoke(
                main.cli,
                [
                    "report",
                    "foo.py",
                    "example_metric",
                    "-n 101",
                    "--message",
                    "--format=CONSOLE",
                ],
            )
            assert result.exit_code == 0, result.stdout
            assert report.called_once
            assert check_cache.called_once
            assert report.call_args[1]["path"] == "foo.py"
            assert report.call_args[1]["metrics"] == ("example_metric",)
            assert report.call_args[1]["include_message"]
            assert report.call_args[1]["n"] == 101
            assert report.call_args[1]["format"] == ReportFormat.CONSOLE


def test_report_not_existing_format():
    """
    Test that report calls the report command with a non-existing format
    """
    with patch("wily.__main__.exists", return_value=True) as check_cache:
        with patch("wily.commands.report.report") as report:
            runner = CliRunner()
            result = runner.invoke(
                main.cli,
                [
                    "report",
                    "foo.py",
                    "example_metric",
                    "-n 101",
                    "--message",
                    "--format=non-existing",
                ],
            )
            assert result.exit_code == 2, result.stdout
            assert report.called_once
            assert check_cache.called_once


def test_report_html_format_with_output():
    """
    Test that report calls the report command with HTML as format and specified output
    """
    with patch("wily.__main__.exists", return_value=True) as check_cache:
        with patch("wily.commands.report.report") as report:
            runner = CliRunner()
            result = runner.invoke(
                main.cli,
                [
                    "report",
                    "foo.py",
                    "example_metric",
                    "-n 101",
                    "--message",
                    "--format=HTML",
                    "--output=reports/out.html",
                ],
            )

            assert result.exit_code == 0, result.stdout
            assert report.called_once
            assert check_cache.called_once
            assert report.call_args[1]["path"] == "foo.py"
            assert report.call_args[1]["metrics"] == ("example_metric",)
            assert report.call_args[1]["include_message"]
            assert report.call_args[1]["n"] == 101
            assert report.call_args[1]["format"] == ReportFormat.HTML
            assert report.call_args[1]["output"] == Path().cwd() / Path(
                "reports/out.html"
            )


def test_graph():
    """
    Test that graph calls the graph command
    """
    with patch("wily.__main__.exists", return_value=True) as check_cache:
        with patch("wily.commands.graph.graph") as graph:
            runner = CliRunner()
            result = runner.invoke(main.cli, ["graph", "foo.py", "example_metric"])
            assert result.exit_code == 0
            assert graph.called_once
            assert check_cache.called_once
            assert graph.call_args[1]["path"] == "foo.py"
            assert graph.call_args[1]["metrics"] == ("example_metric",)


def test_graph_multiple_metrics():
    """
    Test that graph calls the graph command with multiple metrics
    """
    with patch("wily.__main__.exists", return_value=True) as check_cache:
        with patch("wily.commands.graph.graph") as graph:
            runner = CliRunner()
            result = runner.invoke(
                main.cli, ["graph", "foo.py", "example_metric", "another_metric"]
            )
            assert result.exit_code == 0
            assert graph.called_once
            assert check_cache.called_once
            assert graph.call_args[1]["path"] == "foo.py"
            assert graph.call_args[1]["metrics"] == ("example_metric", "another_metric")


def test_graph_with_output():
    """
    Test that graph calls the graph command with output
    """
    with patch("wily.__main__.exists", return_value=True) as check_cache:
        with patch("wily.commands.graph.graph") as graph:
            runner = CliRunner()
            result = runner.invoke(
                main.cli, ["graph", "foo.py", "example_metric", "-o", "foo.html"]
            )
            assert result.exit_code == 0
            assert graph.called_once
            assert check_cache.called_once
            assert graph.call_args[1]["path"] == "foo.py"
            assert graph.call_args[1]["metrics"] == ("example_metric",)
            assert graph.call_args[1]["output"] == "foo.html"


def test_diff():
    """
    Test that diff calls the diff command
    """
    with patch(
        "wily.__main__.get_default_metrics",
        return_value=["maintainability.mi", "raw.loc"],
    ) as gdf:
        with patch("wily.__main__.exists", return_value=True) as check_cache:
            with patch("wily.commands.diff.diff") as diff:
                runner = CliRunner()
                result = runner.invoke(main.cli, ["diff", "foo.py", "x/b.py"])
                assert result.exit_code == 0
                assert diff.called_once
                assert check_cache.called_once
                assert diff.call_args[1]["files"] == ("foo.py", "x/b.py")
                assert gdf.called_once
                assert "maintainability.mi" in diff.call_args[1]["metrics"]


def test_diff_with_metrics():
    """
    Test that diff calls the diff command with additional metrics
    """
    with patch(
        "wily.__main__.get_default_metrics",
        return_value=["maintainability.mi", "raw.loc"],
    ) as gdf:
        with patch("wily.__main__.exists", return_value=True) as check_cache:
            with patch("wily.commands.diff.diff") as diff:
                runner = CliRunner()
                result = runner.invoke(
                    main.cli,
                    [
                        "diff",
                        "foo.py",
                        "x/b.py",
                        "--metrics",
                        "maintainability.mi,raw.sloc",
                    ],
                )
                assert result.exit_code == 0
                assert diff.called_once
                assert check_cache.called_once
                assert diff.call_args[1]["files"] == ("foo.py", "x/b.py")
                assert not gdf.called
                assert "maintainability.mi" in diff.call_args[1]["metrics"]
                assert "raw.loc" not in diff.call_args[1]["metrics"]


def test_clean():
    """
    Test that graph calls the clean command
    """
    with patch("wily.__main__.exists", return_value=True) as check_cache:
        with patch("wily.cache.clean") as clean:
            runner = CliRunner()
            result = runner.invoke(main.cli, ["clean", "--yes"])
            assert result.exit_code == 0
            assert clean.called_once
            assert check_cache.called_once


def test_clean_with_prompt():
    """
    Test that graph calls the clean command
    """
    with patch("wily.__main__.exists", return_value=True) as check_cache:
        with patch("wily.cache.clean") as clean:
            with patch("wily.__main__.input", return_value="y") as mock_input:
                runner = CliRunner()
                result = runner.invoke(main.cli, ["clean"])
                assert result.exit_code == 0
                assert clean.called_once
                assert check_cache.called_once
                assert mock_input.called_once


def test_clean_with_prompt_no_value():
    """
    Test that graph calls the clean command and if enter type "n" it doesn't clean index
    """
    with patch("wily.__main__.exists", return_value=True) as check_cache:
        with patch("wily.cache.clean") as clean:
            with patch("wily.__main__.input", return_value="n") as mock_input:
                runner = CliRunner()
                result = runner.invoke(main.cli, ["clean"])
                assert result.exit_code == 0
                assert not clean.called
                assert check_cache.called_once
                assert mock_input.called_once



"""
Draw graph in HTML for a specific metric.

TODO: Add multiple lines for multiple files
"""
import pathlib

import plotly.graph_objs as go
import plotly.offline

from wily import logger, format_datetime
from wily.operators import resolve_metric, resolve_metric_as_tuple
from wily.state import State


def metric_parts(metric):
    """Convert a metric name into the operator and metric names."""
    operator, met = resolve_metric_as_tuple(metric)
    return operator.name, met.name


def graph(config, path, metrics, output=None, x_axis=None, changes=True, text=False):
    """
    Graph information about the cache and runtime.

    :param config: The configuration.
    :type  config: :class:`wily.config.WilyConfig`

    :param path: The path to the files.
    :type  path: ``list``

    :param metrics: The Y and Z-axis metrics to report on.
    :type  metrics: ``tuple``

    :param output: Save report to specified path instead of opening browser.
    :type  output: ``str``
    """
    logger.debug("Running report command")

    data = []
    state = State(config)
    abs_path = config.path / pathlib.Path(path)

    if x_axis is None:
        x_axis = "history"
    else:
        x_operator, x_key = metric_parts(x_axis)

    if abs_path.is_dir():
        paths = [
            p.relative_to(config.path) for p in pathlib.Path(abs_path).glob("**/*.py")
        ]
    else:
        paths = [path]

    operator, key = metric_parts(metrics[0])
    if len(metrics) == 1:  # only y-axis
        z_axis = None
    else:
        z_axis = resolve_metric(metrics[1])
        z_operator, z_key = metric_parts(metrics[1])
    for path in paths:
        x = []
        y = []
        z = []
        labels = []
        last_y = None
        for rev in state.index[state.default_archiver].revisions:
            labels.append(f"{rev.revision.author_name} <br>{rev.revision.message}")
            try:
                val = rev.get(config, state.default_archiver, operator, str(path), key)
                if val != last_y or not changes:
                    y.append(val)
                    if z_axis:
                        z.append(
                            rev.get(
                                config,
                                state.default_archiver,
                                z_operator,
                                str(path),
                                z_key,
                            )
                        )
                    if x_axis == "history":
                        x.append(format_datetime(rev.revision.date))
                    else:
                        x.append(
                            rev.get(
                                config,
                                state.default_archiver,
                                x_operator,
                                str(path),
                                x_key,
                            )
                        )
                last_y = val
            except KeyError:
                # missing data
                pass

        # Create traces
        trace = go.Scatter(
            x=x,
            y=y,
            mode="lines+markers+text" if text else "lines+markers",
            name=f"{path}",
            ids=state.index[state.default_archiver].revision_keys,
            text=labels,
            marker=dict(
                size=0 if z_axis is None else z,
                color=list(range(len(y))),
                # colorscale='Viridis',
            ),
            xcalendar="gregorian",
            hoveron="points+fills",
        )
        data.append(trace)
    if output:
        filename = output
        auto_open = False
    else:
        filename = "wily-report.html"
        auto_open = True
    y_metric = resolve_metric(metrics[0])
    title = f"{x_axis.capitalize()} of {y_metric.description} for {path}"
    plotly.offline.plot(
        {
            "data": data,
            "layout": go.Layout(
                title=title,
                xaxis={"title": x_axis},
                yaxis={"title": y_metric.description},
            ),
        },
        auto_open=auto_open,
        filename=filename,
    )



import pathlib

import pytest
from mock import patch

import wily.archivers
import wily.archivers.git as git
import wily.config


class MockAuthor(object):
    name = "Mr Test"
    email = "test@test.com"


TEST_AUTHOR = MockAuthor()


class MockCommit(object):
    name_rev = "1234 bbb"
    author = TEST_AUTHOR
    committed_date = "1/1/1990"

    def __init__(self, message):
        self.message = message


class MockRepo(object):
    active_branch = "master"
    bare = False
    _is_dirty = False
    commits = [MockCommit("commit-1"), MockCommit("commit-2")]

    def is_dirty(self):
        return self._is_dirty

    def iter_commits(self, branch, max_count):
        assert branch == self.active_branch
        assert max_count == 99
        return self.commits


@pytest.fixture
def repo(tmpdir):
    repo = MockRepo()
    tmppath = pathlib.Path(tmpdir)
    with open(tmppath / ".gitignore", "w") as test_txt:
        test_txt.write(".wily/")
    repo.path = tmppath
    return repo


def test_basearchiver():
    archiver = wily.archivers.BaseArchiver()
    with pytest.raises(NotImplementedError):
        archiver.revisions("", 10)

    with pytest.raises(NotImplementedError):
        archiver.checkout("")


def test_defaults():
    assert wily.archivers.ARCHIVER_GIT in wily.archivers.ALL_ARCHIVERS.values()


def test_resolve_archiver():
    archiver = wily.archivers.resolve_archiver("git")
    assert archiver == wily.archivers.ARCHIVER_GIT
    assert archiver.name == "git"


def test_bad_resolve_archiver():
    with pytest.raises(ValueError):
        wily.archivers.resolve_archiver("baz")


def test_git_init(repo):
    with patch("wily.archivers.git.Repo", return_value=repo):
        test_config = wily.config.DEFAULT_CONFIG
        test_config.path = repo.path
        archiver = git.GitArchiver(test_config)
        assert archiver.repo is not None
        assert archiver.config == test_config


def test_git_revisions(repo, tmpdir):
    with patch("wily.archivers.git.Repo", return_value=repo):
        test_config = wily.config.DEFAULT_CONFIG
        test_config.path = repo.path
        archiver = git.GitArchiver(test_config)
        revisions = archiver.revisions(tmpdir, 99)
        assert archiver.repo is not None
        assert archiver.config == test_config
        assert revisions[0].key == "1234"
        assert revisions[1].key == "1234"
        assert revisions[0].author_name == "Mr Test"



from click.testing import CliRunner

import wily.__main__ as main


def test_index_no_cache(tmpdir, cache_path):
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", tmpdir, "--cache", cache_path, "index"])
    assert result.exit_code == 1, result.stdout


def test_index(builddir):
    """
    Test that index works with a build
    """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", builddir, "index"])
    assert result.exit_code == 0, result.stdout


def test_index_with_messages(builddir):
    """
    Test that index works with a build with git commit messages
    """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", builddir, "index", "--message"])
    assert result.exit_code == 0, result.stdout



"""
This is really an integration test.
"""

import os.path

import pytest

import wily.config
import wily.state


@pytest.fixture
def config(builddir):
    _cfg = wily.config.DEFAULT_CONFIG
    _cfg.path = builddir
    return _cfg


def test_state_defaults(config):
    """ Test the state defaults """
    state = wily.state.State(config)
    assert state.index
    assert "git" in state.index
    assert state.default_archiver == "git"
    assert state.config is config


def test_index(config):
    """ Test the state index """
    state = wily.state.State(config)
    assert state.index
    assert state.index["git"] is not None
    for revision in state.index["git"].revisions:
        assert state.index["git"][revision.revision.key]
        assert revision.revision in state.index["git"]
        assert revision.revision.key in state.index["git"]



"""
Git Archiver.

Implementation of the archiver API for the gitpython module.
"""
import logging

from git import Repo
import git.exc

from wily.archivers import BaseArchiver, Revision

logger = logging.getLogger(__name__)


class InvalidGitRepositoryError(Exception):
    """Error for when a folder is not a git repo."""

    pass


class DirtyGitRepositoryError(Exception):
    """Error for a dirty git repository (untracked files)."""

    def __init__(self, untracked_files):
        """
        Raise error for untracked files.

        :param untracked_files: List of untracked files
        :param untracked_files: ``list``
        """
        self.untracked_files = untracked_files
        self.message = "Dirty repository, make sure you commit/stash files first"


class GitArchiver(BaseArchiver):
    """Gitpython implementation of the base archiver."""

    name = "git"

    def __init__(self, config):
        """
        Instantiate a new Git Archiver.

        :param config: The wily configuration
        :type  config: :class:`wily.config.WilyConfig`
        """
        try:
            self.repo = Repo(config.path)
        except git.exc.InvalidGitRepositoryError as e:
            raise InvalidGitRepositoryError from e

        self.config = config
        self.current_branch = self.repo.active_branch
        assert not self.repo.bare, "Not a Git repository"

    def revisions(self, path, max_revisions):
        """
        Get the list of revisions.

        :param path: the path to target.
        :type  path: ``str``

        :param max_revisions: the maximum number of revisions.
        :type  max_revisions: ``int``

        :return: A list of revisions.
        :rtype: ``list`` of :class:`Revision`
        """
        if self.repo.is_dirty():
            raise DirtyGitRepositoryError(self.repo.untracked_files)

        revisions = []
        for commit in self.repo.iter_commits(
            self.current_branch, max_count=max_revisions
        ):
            rev = Revision(
                key=commit.name_rev.split(" ")[0],
                author_name=commit.author.name,
                author_email=commit.author.email,
                date=commit.committed_date,
                message=commit.message,
            )
            revisions.append(rev)
        return revisions

    def checkout(self, revision, options):
        """
        Checkout a specific revision.

        :param revision: The revision identifier.
        :type  revision: :class:`Revision`

        :param options: Any additional options.
        :type  options: ``dict``
        """
        rev = revision.key
        self.repo.git.checkout(rev)

    def finish(self):
        """
        Clean up any state if processing completed/failed.

        For git, will checkout HEAD on the original branch when finishing
        """
        self.repo.git.checkout(self.current_branch)
        self.repo.close()



"""
Tests for the cyclomatic complexity operator's ability to handle bad data from radon.
"""
import mock
import wily.operators.cyclomatic
from wily.config import DEFAULT_CONFIG


class MockCC(object):
    results = {}


@mock.patch("wily.operators.cyclomatic.harvesters.CCHarvester", return_value=MockCC)
def test_cyclomatic_bad_entry_data(harvester):
    MockCC.results = {"test.py": [{"complexity": 5}]}
    op = wily.operators.cyclomatic.CyclomaticComplexityOperator(DEFAULT_CONFIG)
    results = op.run("test.py", {})
    assert results == {"test.py": {"complexity": 0}}


@mock.patch("wily.operators.cyclomatic.harvesters.CCHarvester", return_value=MockCC)
def test_cyclomatic_error_case(harvester):
    MockCC.results = {"test.py": {"error": "bad data"}}
    op = wily.operators.cyclomatic.CyclomaticComplexityOperator(DEFAULT_CONFIG)
    results = op.run("test.py", {})
    assert results == {"test.py": {"complexity": 0}}


@mock.patch("wily.operators.cyclomatic.harvesters.CCHarvester", return_value=MockCC)
def test_cyclomatic_error_case_unexpected(harvester):
    MockCC.results = {"test.py": [1234]}
    op = wily.operators.cyclomatic.CyclomaticComplexityOperator(DEFAULT_CONFIG)
    results = op.run("test.py", {})
    assert results == {"test.py": {"complexity": 0}}



"""
For managing the state of the wily process.

Contains a lazy revision, index and process state model.
"""
from collections import OrderedDict
from dataclasses import dataclass, asdict
from typing import List

import wily.cache as cache
from wily import logger
from wily.archivers import Revision, resolve_archiver
from wily.operators import get_metric


@dataclass
class IndexedRevision(object):
    """Union of revision and the operators executed."""

    revision: Revision
    operators: List
    _data = None

    @staticmethod
    def fromdict(d):
        """Instantiate from a dictionary."""
        rev = Revision(
            key=d["key"],
            author_name=d["author_name"],
            author_email=d["author_email"],
            date=d["date"],
            message=d["message"],
        )
        operators = d["operators"]
        return IndexedRevision(revision=rev, operators=operators)

    def asdict(self):
        """Convert to dictionary."""
        d = asdict(self.revision)
        d["operators"] = self.operators
        return d

    def get(self, config, archiver, operator, path, key):
        """
        Get the metric data for this indexed revision.

        :param config: The wily config.
        :type  config: :class:`wily.config.WilyConfig`

        :param archiver: The archiver.
        :type  archiver: :class:`wily.archivers.Archiver`

        :param operator: The operator to find
        :type  operator: ``str``

        :param path: The path to find
        :type  path: ``str``

        :param key: The metric key
        :type  key: ``str``
        """
        if not self._data:
            self._data = cache.get(
                config=config, archiver=archiver, revision=self.revision.key
            )["operator_data"]
        logger.debug(f"Fetching metric {path} - {key} for operator {operator}")
        return get_metric(self._data, operator, path, key)

    def store(self, config, archiver, stats):
        """
        Store the stats for this indexed revision.

        :param config: The wily config.
        :type  config: :class:`wily.config.WilyConfig`

        :param archiver: The archiver.
        :type  archiver: :class:`wily.archivers.Archiver`

        :param stats: The data
        :type  stats: ``dict``
        """
        self._data = stats
        return cache.store(config, archiver, self.revision, stats)


class Index(object):
    """The index of the wily cache."""

    operators = None

    def __init__(self, config, archiver):
        """
        Instantiate a new index.

        :param config: The wily config.
        :type  config: :class:`wily.config.WilyConfig`

        :param archiver: The archiver.
        :type  archiver: :class:`wily.archivers.Archiver`
        """
        self.config = config
        self.archiver = archiver
        self.data = (
            cache.get_archiver_index(config, archiver.name)
            if cache.has_archiver_index(config, archiver.name)
            else []
        )

        self._revisions = OrderedDict()
        for d in self.data:
            self._revisions[d["key"]] = IndexedRevision.fromdict(d)

    def __len__(self):
        """Use length of revisions as len."""
        return len(self._revisions)

    @property
    def revisions(self):
        """
        List of all the revisions.

        :rtype: ``list`` of :class:`LazyRevision`
        """
        return list(self._revisions.values())

    @property
    def revision_keys(self):
        """
        List of all the revision indexes.

        :rtype: ``list`` of ``str``
        """
        return list(self._revisions.keys())

    def __contains__(self, item):
        """
        Check if index contains `item`.

        :param item: The item to search for
        :type  item: ``str``, :class:`Revision` or :class:`LazyRevision`

        :return: ``True`` for contains, ``False`` for not.
        """
        if isinstance(item, Revision):
            return item.key in self._revisions
        elif isinstance(item, str):
            return item in self._revisions
        else:
            raise TypeError("Invalid type for __contains__ in Index.")

    def __getitem__(self, index):
        """Get the revision for a specific index."""
        return self._revisions[index]

    def add(self, revision, operators):
        """
        Add a revision to the index.

        :param revision: The revision.
        :type  revision: :class:`Revision` or :class:`LazyRevision`
        """
        ir = IndexedRevision(
            revision=revision, operators=[operator.name for operator in operators]
        )
        self._revisions[revision.key] = ir
        return ir

    def save(self):
        """Save the index data back to the wily cache."""
        data = [i.asdict() for i in self._revisions.values()]
        logger.debug("Saving data")
        cache.store_archiver_index(self.config, self.archiver, data)


class State(object):
    """
    The wily process state.

    Includes indexes for each archiver.
    """

    def __init__(self, config, archiver=None):
        """
        Instantiate a new process state.

        :param config: The wily configuration.
        :type  config: :class:`WilyConfig`

        :param archiver: The archiver (optional).
        :type  archiver: :class:`wily.archivers.Archiver`
        """
        if archiver:
            self.archivers = [archiver.name]
        else:
            self.archivers = cache.list_archivers(config)
        logger.debug(f"Initialised state indexes for archivers {self.archivers}")
        self.config = config
        self.index = {}
        for archiver in self.archivers:
            self.index[archiver] = Index(self.config, resolve_archiver(archiver))
        self.default_archiver = self.archivers[0]

    def ensure_exists(self):
        """Ensure that cache directory exists."""
        if not cache.exists(self.config):
            logger.debug("Wily cache not found, creating.")
            cache.create(self.config)
            logger.debug("Created wily cache")
        else:
            logger.debug(f"Cache {self.config.cache_path} exists")



"""
Diff command.

Compares metrics between uncommitted files and indexed files.
"""
import os

import tabulate

from wily import logger
from wily.config import DEFAULT_GRID_STYLE
from wily.operators import (
    resolve_metric,
    resolve_operator,
    get_metric,
    GOOD_COLORS,
    BAD_COLORS,
    OperatorLevel,
)
from wily.state import State


def diff(config, files, metrics, changes_only=True, detail=True):
    """
    Show the differences in metrics for each of the files.

    :param config: The wily configuration
    :type  config: :namedtuple:`wily.config.WilyConfig`

    :param files: The files to compare.
    :type  files: ``list`` of ``str``

    :param metrics: The metrics to measure.
    :type  metrics: ``list`` of ``str``

    :param changes_only: Only include changes files in output.
    :type  changes_only: ``bool``

    :param detail: Show details (function-level)
    :type  detail: ``bool``
    """
    config.targets = files
    files = list(files)
    state = State(config)
    last_revision = state.index[state.default_archiver].revisions[0]

    # Convert the list of metrics to a list of metric instances
    operators = {resolve_operator(metric.split(".")[0]) for metric in metrics}
    metrics = [(metric.split(".")[0], resolve_metric(metric)) for metric in metrics]
    data = {}
    results = []

    # Build a set of operators
    _operators = [operator.cls(config) for operator in operators]

    cwd = os.getcwd()
    os.chdir(config.path)
    for operator in _operators:
        logger.debug(f"Running {operator.name} operator")
        data[operator.name] = operator.run(None, config)
    os.chdir(cwd)

    # Write a summary table..
    extra = []
    for operator, metric in metrics:
        if detail and resolve_operator(operator).level == OperatorLevel.Object:
            for file in files:
                try:
                    extra.extend(
                        [
                            f"{file}:{k}"
                            for k in data[operator][file].keys()
                            if k != metric.name
                            and isinstance(data[operator][file][k], dict)
                        ]
                    )
                except KeyError:
                    logger.debug(f"File {file} not in cache")
                    logger.debug("Cache follows -- ")
                    logger.debug(data[operator])
    files.extend(extra)
    logger.debug(files)
    for file in files:
        metrics_data = []
        has_changes = False
        for operator, metric in metrics:
            try:
                current = last_revision.get(
                    config, state.default_archiver, operator, file, metric.name
                )
            except KeyError as e:
                current = "-"
            try:
                new = get_metric(data, operator, file, metric.name)
            except KeyError as e:
                new = "-"
            if new != current:
                has_changes = True
            if metric.type in (int, float) and new != "-" and current != "-":
                if current > new:
                    metrics_data.append(
                        "{0:n} -> \u001b[{2}m{1:n}\u001b[0m".format(
                            current, new, BAD_COLORS[metric.measure]
                        )
                    )
                elif current < new:
                    metrics_data.append(
                        "{0:n} -> \u001b[{2}m{1:n}\u001b[0m".format(
                            current, new, GOOD_COLORS[metric.measure]
                        )
                    )
                else:
                    metrics_data.append("{0:n} -> {1:n}".format(current, new))
            else:
                if current == "-" and new == "-":
                    metrics_data.append("-")
                else:
                    metrics_data.append("{0} -> {1}".format(current, new))
        if has_changes or not changes_only:
            results.append((file, *metrics_data))
        else:
            logger.debug(metrics_data)

    descriptions = [metric.description for operator, metric in metrics]
    headers = ("File", *descriptions)
    if len(results) > 0:
        print(
            # But it still makes more sense to show the newest at the top, so reverse again
            tabulate.tabulate(
                headers=headers, tabular_data=results, tablefmt=DEFAULT_GRID_STYLE
            )
        )



import pathlib
import sys
from textwrap import dedent

from click.testing import CliRunner

import wily.__main__ as main

_path = "src\\test.py" if sys.platform == "win32" else "src/test.py"


def test_diff_no_cache(tmpdir):
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", tmpdir, "diff", _path])
    assert result.exit_code == 1, result.stdout


def test_diff_no_path(tmpdir):
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", tmpdir, "diff"])
    assert result.exit_code == 2, result.stdout


def test_diff_output(builddir):
    """ Test the diff feature with no changes """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--debug", "--path", builddir, "diff", _path])
    assert result.exit_code == 0, result.stdout
    assert "test.py" not in result.stdout


def test_diff_output_all(builddir):
    """ Test the diff feature with no changes and the --all flag """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--debug", "--path", builddir, "diff", _path, "--all"]
    )
    assert result.exit_code == 0, result.stdout
    assert "test.py" in result.stdout


def test_diff_output_bad_path(builddir):
    """ Test the diff feature with no changes """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--debug", "--path", builddir, "diff", "src/baz.py"]
    )
    assert result.exit_code == 0, result.stdout
    assert "test.py" not in result.stdout


def test_diff_output_remove_all(builddir):
    """ Test the diff feature by removing all functions and classes """

    with open(pathlib.Path(builddir) / "src" / "test.py", "w") as test_py:
        test_py.write("print(1)")

    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--debug", "--path", builddir, "diff", _path, "--all"]
    )
    assert result.exit_code == 0, result.stdout


def test_diff_output_more_complex(builddir):
    """ Test the diff feature by making the test file more complicated """

    complex_test = """
            import abc
            foo = 1
            def function1():
                a = 1 + 1
                if a == 2:
                    print(1)
            class Class1(object):
                def method(self):
                    b = 1 + 5
                    if b == 6:
                        if 1==2:
                           if 2==3:
                              print(1)
            """

    with open(pathlib.Path(builddir) / "src" / "test.py", "w") as test_py:
        test_py.write(dedent(complex_test))

    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--debug", "--path", builddir, "diff", _path, "--all"]
    )
    assert result.exit_code == 0, result.stdout
    assert "test.py" in result.stdout
    assert "- -> -" not in result.stdout
    assert "-> -" not in result.stdout
    assert "- ->" not in result.stdout


def test_diff_output_less_complex(builddir):
    """ Test the diff feature by making the test file more complicated """

    simple_test = """
            import abc
            foo = 1
            def function1():
                pass
            class Class1(object):
                def method(self):
                    pass
            """

    with open(pathlib.Path(builddir) / "src" / "test.py", "w") as test_py:
        test_py.write(dedent(simple_test))

    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--debug", "--path", builddir, "diff", _path, "--all"]
    )
    assert result.exit_code == 0, result.stdout
    assert "test.py" in result.stdout
    assert "- -> -" not in result.stdout
    assert "-> -" not in result.stdout
    assert "- ->" not in result.stdout


def test_diff_output_loc(builddir):
    """ Test the diff feature by making the test file more complicated """

    simple_test = """print("test")"""

    with open(pathlib.Path(builddir) / "src" / "test.py", "w") as test_py:
        test_py.write(dedent(simple_test))

    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--debug", "--path", builddir, "diff", _path, "--metrics", "raw.loc"]
    )
    assert result.exit_code == 0, result.stdout
    assert "test.py" in result.stdout
    assert "10 -> \x1b[33m1\x1b[0m" in result.stdout  # 10 -> 1 (in green)


def test_diff_output_rank(builddir):
    """ Test the diff feature by making the test file more complicated """

    simple_test = """print("test")"""

    with open(pathlib.Path(builddir) / "src" / "test.py", "w") as test_py:
        test_py.write(dedent(simple_test))

    runner = CliRunner()
    result = runner.invoke(
        main.cli,
        [
            "--debug",
            "--path",
            builddir,
            "diff",
            _path,
            "--all",
            "--metrics",
            "maintainability.rank",
        ],
    )
    assert result.exit_code == 0, result.stdout
    assert "test.py" in result.stdout
    assert "A -> A" in result.stdout



"""
Maintainability operator.

Measures the "maintainability" using the Halstead index.
"""
import statistics
from collections import Counter

import radon.cli.harvest as harvesters
from radon.cli import Config

from wily import logger
from wily.operators import BaseOperator, MetricType, Metric


def mode(data):
    """
    Return the modal value of a iterable with discrete values.
    
    If there is more than 1 modal value, arbritrarily return the first top n.
    """
    c = Counter(data)
    mode, freq = c.most_common(1)[0]
    return mode


class MaintainabilityIndexOperator(BaseOperator):
    """MI Operator."""

    name = "maintainability"
    defaults = {
        "exclude": None,
        "ignore": None,
        "min": "A",
        "max": "C",
        "multi": True,
        "show": False,
        "sort": False,
    }

    metrics = (
        Metric("rank", "Maintainability Ranking", str, MetricType.Informational, mode),
        Metric(
            "mi", "Maintainability Index", float, MetricType.AimLow, statistics.mean
        ),
    )

    default_metric_index = 1  # MI

    def __init__(self, config):
        """
        Instantiate a new MI operator.

        :param config: The wily configuration.
        :type  config: :class:`WilyConfig`
        """
        # TODO : Import config from wily.cfg
        logger.debug(f"Using {config.targets} with {self.defaults} for MI metrics")

        self.harvester = harvesters.MIHarvester(
            config.targets, config=Config(**self.defaults)
        )

    def run(self, module, options):
        """
        Run the operator.

        :param module: The target module path.
        :type  module: ``str``

        :param options: Any runtime options.
        :type  options: ``dict``

        :return: The operator results.
        :rtype: ``dict``
        """
        logger.debug("Running maintainability harvester")
        return dict(self.harvester.results)



import sys

from click.testing import CliRunner
from pathlib import Path

import wily.__main__ as main

_path = "src\\test.py" if sys.platform == "win32" else "src/test.py"


def test_report_no_cache(tmpdir, cache_path):
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", tmpdir, "--cache", cache_path, "report", _path]
    )
    assert result.exit_code == 1, result.stdout


def test_report(builddir):
    """
    Test that report works with a build and a specific metric
    """
    runner = CliRunner()
    result = runner.invoke(
        main.cli,
        ["--path", builddir, "report", _path, "raw.multi", "maintainability.rank"],
    )
    assert result.exit_code == 0, result.stdout
    assert "Not found" not in result.stdout


def test_report_granular(builddir):
    """
    Test that report works with a build against specific metrics and a function
    """
    runner = CliRunner()
    result = runner.invoke(
        main.cli,
        [
            "--path",
            builddir,
            "report",
            _path + ":function1",
            "cyclomatic.complexity",
            "-n",
            1,
        ],
    )
    assert result.exit_code == 0, result.stdout
    assert "Not found" not in result.stdout


def test_report_not_found(builddir):
    """
    Test that report works with a build but not with an invalid path
    """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", builddir, "report", "test1.py", "raw.loc"]
    )
    assert result.exit_code == 0, result.stdout
    assert "Not found" in result.stdout


def test_report_default_metrics(builddir):
    """
    Test that report works with default metrics
    """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", builddir, "report", _path])
    assert result.exit_code == 0, result.stdout
    assert "Not found" not in result.stdout


def test_report_path(builddir):
    """
    Test that report with a path to a folder (aggregate values)
    """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", builddir, "report", "src"])
    assert result.exit_code == 0, result.stdout
    assert "Not found" not in result.stdout


def test_report_with_message(builddir):
    """
    Test that report works messages in UI
    """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", builddir, "report", _path, "raw.multi", "--message"]
    )
    assert result.exit_code == 0, result.stdout
    assert "basic test" in result.stdout
    assert "remove line" in result.stdout
    assert "Not found" not in result.stdout


def test_report_high_metric(builddir):
    """
    Test that report works with a build on a metric expecting high values
    """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", builddir, "report", _path, "raw.comments"]
    )
    assert result.exit_code == 0, result.stdout
    assert "Not found" not in result.stdout


def test_report_short_metric(builddir):
    """
    Test that report works with a build on shorthand metric
    """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", builddir, "report", _path, "sloc"]
    )
    assert result.exit_code == 0, result.stdout
    assert "Not found" not in result.stdout


def test_report_low_metric(builddir):
    """
    Test that report works with a build on a metric expecting high values
    """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", builddir, "report", _path, "maintainability.mi"]
    )
    assert result.exit_code == 0, result.stdout
    assert "Not found" not in result.stdout


def test_report_html_format(builddir):
    """
    Test that report works with HTML as format
    """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", builddir, "report", _path, "--format", "HTML"]
    )
    path = Path().cwd()
    path = path / "wily_report" / "index.html"

    assert path.exists()
    assert "<html>" in path.read_text()
    assert result.exit_code == 0, result.stdout
    assert "Not found" not in result.stdout


def test_report_html_format_target_folder(builddir):
    """
    Test that report works with HTML as format
    """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", builddir, "report", _path, "--format", "HTML", "-o", "foo"]
    )
    path = Path().cwd()
    path = path / "foo" / "index.html"

    assert path.exists()
    assert "<html>" in path.read_text()
    assert result.exit_code == 0, result.stdout
    assert "Not found" not in result.stdout


def test_report_html_format_target_file(builddir):
    """
    Test that report works with HTML as format
    """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", builddir, "report", _path, "--format", "HTML", "-o", "foo/bar.html"]
    )
    path = Path().cwd()
    path = path / "foo" / "bar.html"

    assert path.exists()
    assert "<html>" in path.read_text()
    assert result.exit_code == 0, result.stdout
    assert "Not found" not in result.stdout


def test_report_console_format(builddir):
    """
    Test that report works with console as format
    """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", builddir, "report", _path, "--format", "CONSOLE"]
    )
    assert result.exit_code == 0, result.stdout
    assert "Not found" not in result.stdout


def test_report_not_existing_format(builddir):
    """
    Test that report works with non-existing format
    """
    runner = CliRunner()
    result = runner.invoke(
        main.cli,
        ["--path", builddir, "report", _path, "--format", "non-existing"],
    )
    assert result.exit_code == 2, result.stdout
    assert "Not found" not in result.stdout



"""Models and types for "operators" the basic measure of a module that measures code."""

from collections import namedtuple
from enum import Enum
from functools import lru_cache


class MetricType(Enum):
    """Type of metric, used in trends."""

    AimLow = 1  # Low is good, high is bad
    AimHigh = 2  # High is good, low is bad
    Informational = 3  # Doesn't matter


Metric = namedtuple("Metric", "name description type measure aggregate")

GOOD_COLORS = {
    MetricType.AimHigh: 32,
    MetricType.AimLow: 31,
    MetricType.Informational: 33,
}

BAD_COLORS = {
    MetricType.AimHigh: 31,
    MetricType.AimLow: 32,
    MetricType.Informational: 33,
}


class OperatorLevel(Enum):
    """Level of operator."""

    File = 1
    Object = 2


class BaseOperator(object):
    """Abstract Operator Class."""

    """Name of the operator."""
    name = "abstract"

    """Default settings."""
    defaults = {}

    """Available metrics as a list of tuple ("name"<str>, "description"<str>, "type"<type>, "metric_type"<MetricType>)."""
    metrics = ()

    """Which metric is the default to display in the report command."""
    default_metric_index = None

    """Level at which the operator goes to."""
    level = OperatorLevel.File

    def run(self, module, options):
        """
        Run the operator.

        :param module: The target module path.
        :type  module: ``str``

        :param options: Any runtime options.
        :type  options: ``dict``

        :return: The operator results.
        :rtype: ``dict``
        """
        raise NotImplementedError


from wily.operators.cyclomatic import CyclomaticComplexityOperator
from wily.operators.maintainability import MaintainabilityIndexOperator
from wily.operators.raw import RawMetricsOperator
from wily.operators.halstead import HalsteadOperator


"""Type for an operator."""
Operator = namedtuple("Operator", "name cls description level")

OPERATOR_CYCLOMATIC = Operator(
    name="cyclomatic",
    cls=CyclomaticComplexityOperator,
    description="Cyclomatic Complexity of modules",
    level=OperatorLevel.Object,
)

OPERATOR_RAW = Operator(
    name="raw",
    cls=RawMetricsOperator,
    description="Raw Python statistics",
    level=OperatorLevel.File,
)

OPERATOR_MAINTAINABILITY = Operator(
    name="maintainability",
    cls=MaintainabilityIndexOperator,
    description="Maintainability index (lines of code and branching)",
    level=OperatorLevel.File,
)

OPERATOR_HALSTEAD = Operator(
    name="halstead",
    cls=HalsteadOperator,
    description="Halstead metrics",
    level=OperatorLevel.Object,
)


"""Dictionary of all operators"""
ALL_OPERATORS = {
    operator.name: operator
    for operator in {
        OPERATOR_CYCLOMATIC,
        OPERATOR_MAINTAINABILITY,
        OPERATOR_RAW,
        OPERATOR_HALSTEAD,
    }
}


"""Set of all metrics"""
ALL_METRICS = {
    (operator, metric)
    for operator in ALL_OPERATORS.values()
    for metric in operator.cls.metrics
}


@lru_cache(maxsize=128)
def resolve_operator(name):
    """
    Get the :namedtuple:`wily.operators.Operator` for a given name.

    :param name: The name of the operator
    :return: The operator type
    """
    if name.lower() in ALL_OPERATORS:
        return ALL_OPERATORS[name.lower()]
    else:
        raise ValueError(f"Operator {name} not recognised.")


def resolve_operators(operators):
    """
    Resolve a list of operator names to their corresponding types.

    :param operators: The list of operators
    :type  operators: iterable or ``str``

    :rtype: ``list`` of :class:`Operator`
    """
    return [resolve_operator(operator) for operator in iter(operators)]


@lru_cache(maxsize=128)
def resolve_metric(metric):
    """
    Resolve metric key to a given target.

    :param metric: the metric name.
    :type  metric: ``str``

    :rtype: :class:`Metric`
    """
    return resolve_metric_as_tuple(metric)[1]


@lru_cache(maxsize=128)
def resolve_metric_as_tuple(metric):
    """
    Resolve metric key to a given target.

    :param metric: the metric name.
    :type  metric: ``str``

    :rtype: :class:`Metric`
    """
    if "." in metric:
        _, metric = metric.split(".")

    r = [
        (operator, match) for operator, match in ALL_METRICS if match[0] == metric
    ]
    if not r or len(r) == 0:
        raise ValueError(f"Metric {metric} not recognised.")
    else:
        return r[0]


def get_metric(revision, operator, path, key):
    """
    Get a metric from the cache.

    :param revision: The revision id.
    :type  revision: ``str``

    :param operator: The operator name.
    :type  operator: ``str``

    :param path: The path to the file/function
    :type  path: ``str``

    :param key: The key of the data
    :type  key: ``str``

    :return: Data from the cache
    :rtype: ``dict``
    """
    if ":" in path:
        part, entry = path.split(":")
        val = revision[operator][part][entry][key]
    else:
        val = revision[operator][path][key]
    return val



"""
Cyclomatic complexity metric for each function/method.

Provided by the radon library.
"""
import statistics

import radon
import radon.cli.harvest as harvesters
from radon.cli import Config
from radon.visitors import Function, Class

from wily import logger
from wily.operators import BaseOperator, Metric, MetricType


class CyclomaticComplexityOperator(BaseOperator):
    """Cyclomatic complexity operator."""

    name = "cyclomatic"
    defaults = {
        "exclude": None,
        "ignore": None,
        "min": "A",
        "max": "F",
        "no_assert": True,
        "show_closures": False,
        "order": radon.complexity.SCORE,
    }

    metrics = (
        Metric(
            "complexity",
            "Cyclomatic Complexity",
            float,
            MetricType.AimLow,
            statistics.mean,
        ),
    )

    default_metric_index = 0  # MI

    def __init__(self, config):
        """
        Instantiate a new Cyclomatic Complexity operator.

        :param config: The wily configuration.
        :type  config: :class:`WilyConfig`
        """
        # TODO: Import config for harvester from .wily.cfg
        logger.debug(f"Using {config.targets} with {self.defaults} for CC metrics")

        self.harvester = harvesters.CCHarvester(
            config.targets, config=Config(**self.defaults)
        )

    def run(self, module, options):
        """
        Run the operator.

        :param module: The target module path.
        :type  module: ``str``

        :param options: Any runtime options.
        :type  options: ``dict``

        :return: The operator results.
        :rtype: ``dict``
        """
        logger.debug("Running CC harvester")
        results = {}
        for filename, details in dict(self.harvester.results).items():
            results[filename] = {}
            total = 0  # running CC total
            for instance in details:
                if isinstance(instance, Class):
                    i = self._dict_from_class(instance)
                elif isinstance(instance, Function):
                    i = self._dict_from_function(instance)
                else:
                    if isinstance(instance, str) and instance == "error":
                        logger.warning(
                            f"Failed to run CC harvester on {filename} : {details['error']}"
                        )
                        continue
                    else:
                        logger.warning(
                            f"Unexpected result from Radon : {instance} of {type(instance)}. Please report on Github."
                        )
                        continue
                results[filename][i["fullname"]] = i
                del i["fullname"]
                total += i["complexity"]
            results[filename]["complexity"] = total
        return results

    @staticmethod
    def _dict_from_function(l):
        return {
            "name": l.name,
            "is_method": l.is_method,
            "classname": l.classname,
            "closures": l.closures,
            "complexity": l.complexity,
            "fullname": l.fullname,
            "loc": l.endline - l.lineno,
        }

    @staticmethod
    def _dict_from_class(l):
        return {
            "name": l.name,
            "inner_classes": l.inner_classes,
            "real_complexity": l.real_complexity,
            "complexity": l.complexity,
            "fullname": l.fullname,
            "loc": l.endline - l.lineno,
        }



import pytest
from mock import patch

import wily.commands.build as build
from wily.config import DEFAULT_CONFIG
from wily.archivers import Archiver, BaseArchiver, Revision
from wily.operators import Operator, OperatorLevel, BaseOperator


class MockArchiverCls(BaseArchiver):
    name = "test"

    def __init__(self, *args, **kwargs):
        pass

    def revisions(self, path, max_revisions):
        return [
            Revision(
                key="12345",
                author_name="Local User",  # Don't want to leak local data
                author_email="-",  # as above
                date=12_345_679,
                message="None",
            )
        ]

    def checkout(self, revision, options):
        pass  # noop


class MockOperatorCls(BaseOperator):
    name = "test"
    data = {}

    def __init__(self, *args, **kwargs):
        pass

    def run(self, module, options):
        return self.data


MockOperator = Operator("mock", MockOperatorCls, "for testing", OperatorLevel.File)

MockArchiver = Archiver("mock", MockArchiverCls, "for testing")


@pytest.fixture
def config():
    cfg = DEFAULT_CONFIG
    return cfg


def test_build_simple(config):
    _test_operators = (MockOperator,)
    with patch("wily.state.resolve_archiver", return_value=MockArchiver):
        result = build.build(config, MockArchiver, _test_operators)
    assert result is None



"""
Raw statistics operator.

Includes insights like lines-of-code, number of comments. Does not measure complexity.
"""
import radon.cli.harvest as harvesters
from radon.cli import Config

from wily import logger
from wily.operators import BaseOperator, MetricType, Metric


class RawMetricsOperator(BaseOperator):
    """Raw Metrics Operator."""

    name = "raw"
    defaults = {"exclude": None, "ignore": None, "summary": False}
    metrics = (
        Metric("loc", "Lines of Code", int, MetricType.Informational, sum),
        Metric("lloc", "L Lines of Code", int, MetricType.AimLow, sum),
        Metric("sloc", "S Lines of Code", int, MetricType.AimLow, sum),
        Metric("comments", "Multi-line comments", int, MetricType.AimHigh, sum),
        Metric("multi", "Multi lines", int, MetricType.Informational, sum),
        Metric("blank", "blank lines", int, MetricType.Informational, sum),
        Metric(
            "single_comments",
            "Single comment lines",
            int,
            MetricType.Informational,
            sum,
        ),
    )
    default_metric_index = 0  # LOC

    def __init__(self, config):
        """
        Instantiate a new raw operator.

        :param config: The wily configuration.
        :type  config: :class:`WilyConfig`
        """
        # TODO: Use config from wily.cfg for harvester
        logger.debug(f"Using {config.targets} with {self.defaults} for Raw metrics")
        self.harvester = harvesters.RawHarvester(
            config.targets, config=Config(**self.defaults)
        )

    def run(self, module, options):
        """
        Run the operator.

        :param module: The target module path.
        :type  module: ``str``

        :param options: Any runtime options.
        :type  options: ``dict``

        :return: The operator results.
        :rtype: ``dict``
        """
        logger.debug("Running raw harvester")
        return dict(self.harvester.results)



"""
Archivers module.

Specifies a standard interface for finding revisions (versions) of a path and switching to them.
"""

from collections import namedtuple
from dataclasses import dataclass


class BaseArchiver(object):
    """Abstract Archiver Class."""

    def revisions(self, path, max_revisions):
        """
        Get the list of revisions.

        :param path: the path to target.
        :type  path: ``str``

        :param max_revisions: the maximum number of revisions.
        :type  max_revisions: ``int``

        :return: A list of revisions.
        :rtype: ``list`` of :class:`Revision`
        """
        raise NotImplementedError

    def checkout(self, revision, **options):
        """
        Checkout a specific revision.

        :param revision: The revision identifier.
        :type  revision: :class:`Revision`

        :param options: Any additional options.
        :type  options: ``dict``
        """
        raise NotImplementedError

    def finish(self):
        """Clean up any state if processing completed/failed."""
        pass


@dataclass
class Revision:
    """Represents a revision in the archiver."""

    key: str
    author_name: str
    author_email: str
    date: str
    message: str


from wily.archivers.git import GitArchiver
from wily.archivers.filesystem import FilesystemArchiver


"""Type for an operator"""
Archiver = namedtuple("Archiver", "name cls description")


"""Git Operator defined in `wily.archivers.git`"""
ARCHIVER_GIT = Archiver(name="git", cls=GitArchiver, description="Git archiver")

"""Filesystem archiver"""
ARCHIVER_FILESYSTEM = Archiver(
    name="filesystem", cls=FilesystemArchiver, description="Filesystem archiver"
)

"""Set of all available archivers"""
ALL_ARCHIVERS = {a.name: a for a in [ARCHIVER_GIT, ARCHIVER_FILESYSTEM]}


def resolve_archiver(name):
    """
    Get the :class:`wily.archivers.Archiver` for a given name.

    :param name: The name of the archiver
    :type  name: ``str``
    :return: The archiver type
    """
    if name not in ALL_ARCHIVERS:
        raise ValueError(f"Resolver {name} not recognised.")
    else:
        return ALL_ARCHIVERS[name.lower()]



"""
A module including decorators for wily.

This API is not intended to be public and should not be consumed directly.

"""

from wily import __version__


def add_version(f):
    """
    Add the version of wily to the help heading.

    :param f: function to decorate
    :return: decorated function
    """
    doc = f.__doc__
    f.__doc__ = "Version: " + __version__ + "\n\n" + doc
    return f



import os.path

import wily.config


def test_config_empty_defaults(tmpdir):
    """
    Test that an empty config path sets to defaults.
    """
    config = """
    """
    config_path = os.path.join(tmpdir, "wily.cfg")
    with open(config_path, "w") as config_f:
        config_f.write(config)

    cfg = wily.config.load(config_path)

    assert cfg.archiver == wily.config.DEFAULT_ARCHIVER
    assert cfg.operators == wily.config.DEFAULT_OPERATORS
    assert cfg.max_revisions == wily.config.DEFAULT_MAX_REVISIONS


def test_config_archiver(tmpdir):
    """
    Test that an archiver can be configured
    """
    config = """
    [wily]
    archiver = foo
    """
    config_path = os.path.join(tmpdir, "wily.cfg")
    with open(config_path, "w") as config_f:
        config_f.write(config)

    cfg = wily.config.load(config_path)

    assert cfg.archiver == "foo"
    assert cfg.operators == wily.config.DEFAULT_OPERATORS
    assert cfg.max_revisions == wily.config.DEFAULT_MAX_REVISIONS


def test_config_operators(tmpdir):
    """
    Test that operators can be configured
    """
    config = """
    [wily]
    operators = foo,baz
    """
    config_path = os.path.join(tmpdir, "wily.cfg")
    with open(config_path, "w") as config_f:
        config_f.write(config)

    cfg = wily.config.load(config_path)

    assert cfg.archiver == wily.config.DEFAULT_ARCHIVER
    assert cfg.operators == "foo,baz"
    assert cfg.max_revisions == wily.config.DEFAULT_MAX_REVISIONS


def test_config_archiver(tmpdir):
    """
    Test that an max-revisions can be configured
    """
    config = """
    [wily]
    max_revisions = 14
    """
    config_path = os.path.join(tmpdir, "wily.cfg")
    with open(config_path, "w") as config_f:
        config_f.write(config)

    cfg = wily.config.load(config_path)

    assert cfg.archiver == wily.config.DEFAULT_ARCHIVER
    assert cfg.operators == wily.config.DEFAULT_OPERATORS
    assert cfg.max_revisions == 14



"""
Configuration of wily.

TODO : Handle operator settings. Maybe a section for each operator and then pass kwargs to operators?
TODO : Better utilise default values and factory in @dataclass to replace DEFAULT_CONFIG
 and replace the logic in load() to set default values.
"""
from functools import lru_cache
import configparser
import logging
import pathlib
import hashlib
from dataclasses import dataclass, field
from typing import Any, List

import wily.operators as operators
from wily.archivers import ARCHIVER_GIT

logger = logging.getLogger(__name__)


@lru_cache(maxsize=128)
def generate_cache_path(path):
    """
    Generate a reusable path to cache results.

    Will use the --path of the target and hash into
    a 9-character directory within the HOME folder.

    :return: The cache path
    :rtype: ``str``
    """
    logger.debug(f"Generating cache for {path}")
    sha = hashlib.sha1(str(path).encode()).hexdigest()[:9]
    HOME = pathlib.Path.home()
    cache_path = str(HOME / ".wily" / sha)
    logger.debug(f"Cache path is {cache_path}")
    return cache_path


@dataclass
class WilyConfig(object):
    """
    Wily configuration.

    A data class to reflect the configurable options within Wily.
    """

    operators: List
    archiver: Any
    path: str
    max_revisions: int
    targets: List[str] = None
    checkout_options: dict = field(default_factory=dict)

    def __post_init__(self):
        """Clone targets as a list of path."""
        if self.targets is None or "":
            self.targets = [self.path]
        self._cache_path = None

    @property
    def cache_path(self):
        """Path to the cache."""
        if not self._cache_path:
            self._cache_path = generate_cache_path(pathlib.Path(self.path).absolute())
        return self._cache_path

    @cache_path.setter
    def cache_path(self, value):
        """Override the cache path."""
        logger.debug(f"Setting custom cache path to {value}")
        self._cache_path = value


# Default values for Wily

""" The default operators """
DEFAULT_OPERATORS = {
    operators.OPERATOR_RAW.name,
    operators.OPERATOR_MAINTAINABILITY.name,
    operators.OPERATOR_CYCLOMATIC.name,
    operators.OPERATOR_HALSTEAD.name,
}

""" The name of the default archiver """
DEFAULT_ARCHIVER = ARCHIVER_GIT.name

""" The default configuration file name """
DEFAULT_CONFIG_PATH = "wily.cfg"

""" The default section name in the config """
DEFAULT_CONFIG_SECTION = "wily"

""" The default maximum number of revisions to archiver """
DEFAULT_MAX_REVISIONS = 50

DEFAULT_PATH = "."

""" The default configuration for Wily (if no config file exists) """
DEFAULT_CONFIG = WilyConfig(
    operators=DEFAULT_OPERATORS,
    archiver=DEFAULT_ARCHIVER,
    path=DEFAULT_PATH,
    max_revisions=DEFAULT_MAX_REVISIONS,
)

""" Default table style in console. See tabulate docs for more. """
DEFAULT_GRID_STYLE = "fancy_grid"


def load(config_path=DEFAULT_CONFIG_PATH):
    """
    Load config file and set values to defaults where no present.

    :return: The configuration ``WilyConfig``
    :rtype: :class:`wily.config.WilyConfig`
    """
    if not pathlib.Path(config_path).exists():
        logger.debug(f"Could not locate {config_path}, using default config.")
        return DEFAULT_CONFIG

    config = configparser.ConfigParser(default_section=DEFAULT_CONFIG_SECTION)
    config.read(config_path)

    operators = config.get(
        section=DEFAULT_CONFIG_SECTION, option="operators", fallback=DEFAULT_OPERATORS
    )
    archiver = config.get(
        section=DEFAULT_CONFIG_SECTION, option="archiver", fallback=DEFAULT_ARCHIVER
    )
    path = config.get(section=DEFAULT_CONFIG_SECTION, option="path", fallback=".")
    max_revisions = int(
        config.get(
            section=DEFAULT_CONFIG_SECTION,
            option="max_revisions",
            fallback=DEFAULT_MAX_REVISIONS,
        )
    )

    return WilyConfig(
        operators=operators, archiver=archiver, path=path, max_revisions=max_revisions
    )



"""
Builds a cache based on a source-control history.

TODO : Convert .gitignore to radon ignore patterns to make the build more efficient.

"""
import pathlib
import multiprocessing
from progress.bar import Bar

from wily import logger
from wily.state import State

from wily.archivers.git import InvalidGitRepositoryError
from wily.archivers import FilesystemArchiver

from wily.operators import resolve_operator


def run_operator(operator, revision, config):
    """Run an operator for the multiprocessing pool. Not called directly."""
    instance = operator.cls(config)
    logger.debug(f"Running {operator.name} operator on {revision.key}")
    return operator.name, instance.run(revision, config)


def build(config, archiver, operators):
    """
    Build the history given a archiver and collection of operators.

    :param config: The wily configuration
    :type  config: :namedtuple:`wily.config.WilyConfig`

    :param archiver: The archiver to use
    :type  archiver: :namedtuple:`wily.archivers.Archiver`

    :param operators: The list of operators to execute
    :type operators: `list` of :namedtuple:`wily.operators.Operator`
    """
    try:
        logger.debug(f"Using {archiver.name} archiver module")
        archiver = archiver.cls(config)
        revisions = archiver.revisions(config.path, config.max_revisions)
    except InvalidGitRepositoryError:
        # TODO: This logic shouldn't really be here (SoC)
        logger.info(f"Defaulting back to the filesystem archiver, not a valid git repo")
        archiver = FilesystemArchiver(config)
        revisions = archiver.revisions(config.path, config.max_revisions)
    except Exception as e:
        if hasattr(e, "message"):
            logger.error(f"Failed to setup archiver: '{e.message}'")
        else:
            logger.error(f"Failed to setup archiver: '{type(e)} - {e}'")
        exit(1)

    state = State(config, archiver=archiver)
    # Check for existence of cache, else provision
    state.ensure_exists()

    index = state.index[archiver.name]

    # remove existing revisions from the list
    revisions = [revision for revision in revisions if revision not in index]

    logger.info(
        f"Found {len(revisions)} revisions from '{archiver.name}' archiver in '{config.path}'."
    )

    _op_desc = ",".join([operator.name for operator in operators])
    logger.info(f"Running operators - {_op_desc}")

    bar = Bar("Processing", max=len(revisions) * len(operators))
    state.operators = operators
    try:
        with multiprocessing.Pool(processes=len(operators)) as pool:
            for revision in revisions:
                # Checkout target revision
                archiver.checkout(revision, config.checkout_options)
                stats = {"operator_data": {}}

                # Run each operator as a seperate process
                data = pool.starmap(
                    run_operator,
                    [(operator, revision, config) for operator in operators],
                )

                # Map the data back into a dictionary
                for operator_name, result in data:
                    # aggregate values to directories
                    roots = []

                    # find all unique directories in the results
                    for entry in result.keys():
                        parent = pathlib.Path(entry).parents[0]
                        if parent not in roots:
                            roots.append(parent)

                    for root in roots:
                        # find all matching entries recursively
                        aggregates = [
                            path
                            for path in result.keys()
                            if root in pathlib.Path(path).parents
                        ]
                        result[str(root)] = {}
                        # aggregate values
                        for metric in resolve_operator(operator_name).cls.metrics:
                            func = metric.aggregate
                            values = [
                                result[aggregate][metric.name]
                                for aggregate in aggregates
                                if aggregate in result
                                and metric.name in result[aggregate]
                            ]
                            if len(values) > 0:
                                result[str(root)][metric.name] = func(values)

                    stats["operator_data"][operator_name] = result
                    bar.next()

                ir = index.add(revision, operators=operators)
                ir.store(config, archiver, stats)
        index.save()
        bar.finish()
    except Exception as e:
        logger.error(f"Failed to build cache: '{e}'")
        raise e
    finally:
        # Reset the archive after every run back to the head of the branch
        archiver.finish()



"""
Tests for the wily build command.

All of the following tests will use a click CLI runner to fully simulate the CLI.
Many of the tests will depend on a "builddir" fixture which is a compiled wily cache.

TODO : Test build + build with extra operator
"""
import sys
import pathlib
import pytest
from click.testing import CliRunner
from git import Repo, Actor
from mock import patch

import wily.__main__ as main
from wily.archivers import ALL_ARCHIVERS
from wily.config import generate_cache_path

_path = "src\\test.py" if sys.platform == "win32" else "src/test.py"


def test_build_not_git_repo(tmpdir, cache_path):
    """
    Test that build defaults to filesystem in a non-git directory
    """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", tmpdir, "--cache", cache_path, "build", "test.py"]
    )
    assert result.exit_code == 0, result.stdout
    cache_path = pathlib.Path(cache_path)
    assert cache_path.exists()
    index_path = cache_path / "filesystem" / "index.json"
    assert index_path.exists()


def test_build_custom_cache(tmpdir):
    """
    Test that build defaults to filesystem in a non-git directory with custom cache path.
    """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", tmpdir, "--cache", tmpdir / ".wily", "build", "test.py"]
    )
    assert result.exit_code == 0, result.stdout
    cache_path = tmpdir / ".wily"
    assert cache_path.exists()
    index_path = cache_path / "filesystem" / "index.json"
    assert index_path.exists()
    assert not pathlib.Path(generate_cache_path(tmpdir)).exists()


def test_build_invalid_path():
    """
    Test that build fails with a garbage path
    """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", "/fo/v/a", "build", "test.py"])
    assert result.exit_code == 1, result.stdout


def test_build_crash(tmpdir):
    """
    Simulate a runtime error in the build.
    """
    repo = Repo.init(path=tmpdir)
    tmppath = pathlib.Path(tmpdir)

    # Write a test file to the repo
    with open(tmppath / "test.py", "w") as test_txt:
        test_txt.write("import abc")

    index = repo.index
    index.add(["test.py"])

    author = Actor("An author", "author@example.com")
    committer = Actor("A committer", "committer@example.com")

    index.commit("basic test", author=author, committer=committer)
    repo.close()

    import wily.commands.build

    with patch.object(
        wily.commands.build.Bar, "finish", side_effect=RuntimeError("arggh")
    ) as bar_finish:
        runner = CliRunner()
        result = runner.invoke(main.cli, ["--path", tmpdir, "build", "test.py"])
        assert bar_finish.called_once
        assert result.exit_code == 1, result.stdout


def test_build(tmpdir, cache_path):
    """
    Test that build works in a basic repository.
    """
    repo = Repo.init(path=tmpdir)
    tmppath = pathlib.Path(tmpdir)

    # Write a test file to the repo
    with open(tmppath / "test.py", "w") as test_txt:
        test_txt.write("import abc")

    index = repo.index
    index.add(["test.py"])

    author = Actor("An author", "author@example.com")
    committer = Actor("A committer", "committer@example.com")

    commit = index.commit("basic test", author=author, committer=committer)
    repo.close()

    runner = CliRunner()
    result = runner.invoke(
        main.cli,
        ["--debug", "--path", tmpdir, "--cache", cache_path, "build", "test.py"],
    )
    assert result.exit_code == 0, result.stdout

    cache_path = pathlib.Path(cache_path)
    assert cache_path.exists()
    index_path = cache_path / "git" / "index.json"
    assert index_path.exists()
    rev_path = cache_path / "git" / (commit.name_rev.split(" ")[0] + ".json")
    assert rev_path.exists()


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
def test_build_with_config(tmpdir, cache_path):
    """
    Test that build works in a basic repository and a configuration file.
    """
    repo = Repo.init(path=tmpdir)
    tmppath = pathlib.Path(tmpdir)

    config = """
    [wily]
    path = test.py
    """
    config_path = tmppath / "wily.cfg"
    with open(config_path, "w") as config_f:
        config_f.write(config)

    # Write a test file to the repo
    with open(tmppath / "test.py", "w") as test_txt:
        test_txt.write("import abc")

    index = repo.index
    index.add(["test.py"])

    author = Actor("An author", "author@example.com")
    committer = Actor("A committer", "committer@example.com")

    commit = index.commit("basic test", author=author, committer=committer)
    repo.close()

    runner = CliRunner()
    result = runner.invoke(
        main.cli,
        ["--debug", "--config", config_path, "--path", tmpdir, "--cache", cache_path, "build"],
    )
    assert result.exit_code == 0, result.stdout

    cache_path = pathlib.Path(cache_path)
    assert cache_path.exists()
    index_path = cache_path / "git" / "index.json"
    assert index_path.exists()
    rev_path = cache_path / "git" / (commit.name_rev.split(" ")[0] + ".json")
    assert rev_path.exists()


def test_build_twice(tmpdir, cache_path):
    """
    Test that build works when run twice.
    """
    repo = Repo.init(path=tmpdir)
    tmppath = pathlib.Path(tmpdir)

    # Write a test file to the repo
    with open(tmppath / "test.py", "w") as test_txt:
        test_txt.write("import abc")

    index = repo.index
    index.add(["test.py"])

    author = Actor("An author", "author@example.com")
    committer = Actor("A committer", "committer@example.com")

    commit = index.commit("basic test", author=author, committer=committer)

    runner = CliRunner()
    result = runner.invoke(
        main.cli,
        ["--debug", "--path", tmpdir, "--cache", cache_path, "build", "test.py"],
    )
    assert result.exit_code == 0, result.stdout

    cache_path = pathlib.Path(cache_path) / "git"
    assert cache_path.exists()
    index_path = cache_path / "index.json"
    assert index_path.exists()
    rev_path = cache_path / (commit.name_rev.split(" ")[0] + ".json")
    assert rev_path.exists()

    # Write a test file to the repo
    with open(tmppath / "test.py", "w") as test_txt:
        test_txt.write("import abc\nfoo = 1")

    index.add(["test.py"])

    commit2 = index.commit("basic test", author=author, committer=committer)
    repo.close()

    result = runner.invoke(main.cli, ["--debug", "--path", tmpdir, "build", "test.py"])
    assert result.exit_code == 0, result.stdout

    assert cache_path.exists()
    index_path = cache_path / "index.json"
    assert index_path.exists()
    rev_path = cache_path / (commit.name_rev.split(" ")[0] + ".json")
    assert rev_path.exists()
    rev_path2 = cache_path / (commit2.name_rev.split(" ")[0] + ".json")
    assert rev_path2.exists()


def test_build_no_commits(tmpdir):
    """
    Test that build fails cleanly with no commits
    """
    repo = Repo.init(path=tmpdir)
    repo.close()

    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--debug", "--path", tmpdir, "build", tmpdir, "--skip-ignore-check"]
    )
    assert result.exit_code == 1, result.stdout


def test_build_dirty_repo(builddir):
    """
    Test that build fails cleanly with a dirty repo
    """
    tmppath = pathlib.Path(builddir)
    with open(tmppath / "test.py", "w") as test_txt:
        test_txt.write("import abc\nfoo = 1")

    runner = CliRunner()
    result = runner.invoke(main.cli, ["--debug", "--path", builddir, "build", builddir])
    assert result.exit_code == 1, result.stdout


def test_build_no_git_history(tmpdir):
    repo = Repo.init(path=tmpdir)
    repo.close()

    with patch("wily.logger") as logger:
        runner = CliRunner()
        result = runner.invoke(main.cli, ["--path", tmpdir, "build", _path])
        assert result.exit_code == 1, result.stdout


archivers = {name for name in ALL_ARCHIVERS.keys()}


@pytest.mark.parametrize("archiver", archivers)
def test_build_archiver(gitdir, archiver, cache_path):
    """
    Test the build against each type of archiver
    """
    with patch("wily.logger") as logger:
        runner = CliRunner()
        result = runner.invoke(
            main.cli,
            ["--path", gitdir, "--cache", cache_path, "build", _path, "-a", archiver],
        )
        assert result.exit_code == 0, result.stdout
        cache_path = pathlib.Path(cache_path)
        assert cache_path.exists()
        index_path = cache_path / archiver / "index.json"
        assert index_path.exists()



"""
Report command.

The report command gives a table of metrics for a specified list of files.
Will compare the values between revisions and highlight changes in green/red.
"""
import tabulate

from pathlib import Path
from shutil import copytree
from string import Template

from wily import logger, format_date, format_revision, MAX_MESSAGE_WIDTH
from wily.helper.custom_enums import ReportFormat
from wily.operators import resolve_metric_as_tuple, MetricType
from wily.state import State


def report(
    config, path, metrics, n, output, include_message=False, format=ReportFormat.CONSOLE, console_format=None,
):
    """
    Show information about the cache and runtime.

    :param config: The configuration
    :type  config: :class:`wily.config.WilyConfig`

    :param path: The path to the file
    :type  path: ``str``

    :param metrics: Name of the metric to report on
    :type  metrics: ``str``

    :param n: Number of items to list
    :type  n: ``int``

    :param output: Output path
    :type  output: ``Path``

    :param include_message: Include revision messages
    :type  include_message: ``bool``

    :param format: Output format
    :type  format: ``ReportFormat``

    :param console_format: Grid format style for tabulate
    :type  console_format: ``str``
    """
    logger.debug("Running report command")
    logger.info(f"-----------History for {metrics}------------")

    data = []
    metric_metas = []

    for metric in metrics:
        operator, metric = resolve_metric_as_tuple(metric)
        key = metric.name
        operator = operator.name
        # Set the delta colors depending on the metric type
        if metric.measure == MetricType.AimHigh:
            good_color = 32
            bad_color = 31
        elif metric.measure == MetricType.AimLow:
            good_color = 31
            bad_color = 32
        elif metric.measure == MetricType.Informational:
            good_color = 33
            bad_color = 33
        metric_meta = {
            "key": key,
            "operator": operator,
            "good_color": good_color,
            "bad_color": bad_color,
            "title": metric.description,
            "type": metric.type,
        }
        metric_metas.append(metric_meta)

    state = State(config)
    for archiver in state.archivers:
        # We have to do it backwards to get the deltas between releases
        history = state.index[archiver].revisions[:n][::-1]
        last = {}
        for rev in history:
            vals = []
            for meta in metric_metas:
                try:
                    logger.debug(
                        f"Fetching metric {meta['key']} for {meta['operator']} in {path}"
                    )
                    val = rev.get(config, archiver, meta["operator"], path, meta["key"])

                    last_val = last.get(meta["key"], None)
                    # Measure the difference between this value and the last
                    if meta["type"] in (int, float):
                        if last_val:
                            delta = val - last_val
                        else:
                            delta = 0
                        last[meta["key"]] = val
                    else:
                        # TODO : Measure ranking increases/decreases for str types?
                        delta = 0

                    if delta == 0:
                        delta_col = delta
                    elif delta < 0:
                        delta_col = f"\u001b[{meta['good_color']}m{delta:n}\u001b[0m"
                    else:
                        delta_col = f"\u001b[{meta['bad_color']}m+{delta:n}\u001b[0m"

                    if meta["type"] in (int, float):
                        k = f"{val:n} ({delta_col})"
                    else:
                        k = f"{val}"
                except KeyError as e:
                    k = f"Not found {e}"
                vals.append(k)
            if include_message:
                data.append(
                    (
                        format_revision(rev.revision.key),
                        rev.revision.message[:MAX_MESSAGE_WIDTH],
                        rev.revision.author_name,
                        format_date(rev.revision.date),
                        *vals,
                    )
                )
            else:
                data.append(
                    (
                        format_revision(rev.revision.key),
                        rev.revision.author_name,
                        format_date(rev.revision.date),
                        *vals,
                    )
                )
    descriptions = [meta["title"] for meta in metric_metas]
    if include_message:
        headers = ("Revision", "Message", "Author", "Date", *descriptions)
    else:
        headers = ("Revision", "Author", "Date", *descriptions)

    if format == ReportFormat.HTML:
        if output.is_file and output.suffix == ".html":
            report_path = output.parents[0]
            report_output = output
        else:
            report_path = output
            report_output = output.joinpath("index.html")

        report_path.mkdir(exist_ok=True, parents=True)

        templates_dir = (Path(__file__).parents[1] / "templates").resolve()
        report_template = Template((templates_dir / "report_template.html").read_text())

        table_headers = "".join([f"<th>{header}</th>" for header in headers])
        table_content = ""
        for line in data[::-1]:
            table_content += "<tr>"
            for element in line:
                element = element.replace("[32m", "<span class='green-color'>")
                element = element.replace("[31m", "<span class='red-color'>")
                element = element.replace("[33m", "<span class='orange-color'>")
                element = element.replace("[0m", "</span>")
                table_content += f"<td>{element}</td>"
            table_content += "</tr>"

        report_template = report_template.safe_substitute(
            headers=table_headers, content=table_content
        )

        with report_output.open("w") as output:
            output.write(report_template)

        try:
            copytree(str(templates_dir / "css"), str(report_path / "css"))
        except FileExistsError:
            pass

        logger.info(f"wily report was saved to {report_path}")
    else:
        print(
            # But it still makes more sense to show the newest at the top, so reverse again
            tabulate.tabulate(
                headers=headers, tabular_data=data[::-1], tablefmt=console_format
            )
        )



"""
Print command.

Print information about the wily cache and what is in the index.
"""
import tabulate

from wily import logger, format_date, format_revision, MAX_MESSAGE_WIDTH
from wily.config import DEFAULT_GRID_STYLE
from wily.state import State


def index(config, include_message=False):
    """
    Show information about the cache and runtime.

    :param config: The wily configuration
    :type  config: :namedtuple:`wily.config.WilyConfig`

    :param include_message: Include revision messages
    :type  include_message: ``bool``
    """
    state = State(config=config)
    logger.debug("Running show command")
    logger.info("--------Configuration---------")
    logger.info(f"Path: {config.path}")
    logger.info(f"Archiver: {config.archiver}")
    logger.info(f"Operators: {config.operators}")
    logger.info("")
    logger.info("-----------History------------")

    data = []
    for archiver in state.archivers:
        for rev in state.index[archiver].revisions:
            if include_message:
                data.append(
                    (
                        format_revision(rev.revision.key),
                        rev.revision.author_name,
                        rev.revision.message[:MAX_MESSAGE_WIDTH],
                        format_date(rev.revision.date),
                    )
                )
            else:
                data.append(
                    (
                        format_revision(rev.revision.key),
                        rev.revision.author_name,
                        format_date(rev.revision.date),
                    )
                )

    if include_message:
        headers = ("Revision", "Author", "Message", "Date")
    else:
        headers = ("Revision", "Author", "Date")
    print(
        tabulate.tabulate(
            headers=headers, tabular_data=data, tablefmt=DEFAULT_GRID_STYLE
        )
    )



"""
Halstead operator.

Measures all of the halstead metrics (volume, vocab, difficulty)
"""
import radon.cli.harvest as harvesters
from radon.cli import Config

from wily import logger
from wily.operators import BaseOperator, MetricType, Metric


class HalsteadOperator(BaseOperator):
    """Halstead Operator."""

    name = "halstead"
    defaults = {
        "exclude": None,
        "ignore": None,
        "min": "A",
        "max": "C",
        "multi": True,
        "show": False,
        "sort": False,
        "by_function": True,
    }

    metrics = (
        Metric("h1", "Unique Operands", int, MetricType.AimLow, sum),
        Metric("h2", "Unique Operators", int, MetricType.AimLow, sum),
        Metric("N1", "Number of Operands", int, MetricType.AimLow, sum),
        Metric("N2", "Number of Operators", int, MetricType.AimLow, sum),
        Metric(
            "vocabulary", "Unique vocabulary (h1 + h2)", int, MetricType.AimLow, sum
        ),
        Metric("length", "Length of application", int, MetricType.AimLow, sum),
        Metric("volume", "Code volume", float, MetricType.AimLow, sum),
        Metric("difficulty", "Difficulty", float, MetricType.AimLow, sum),
        Metric("effort", "Effort", float, MetricType.AimLow, sum),
    )

    default_metric_index = 0  # MI

    def __init__(self, config):
        """
        Instantiate a new HC operator.

        :param config: The wily configuration.
        :type  config: :class:`WilyConfig`
        """
        # TODO : Import config from wily.cfg
        logger.debug(f"Using {config.targets} with {self.defaults} for HC metrics")

        self.harvester = harvesters.HCHarvester(
            config.targets, config=Config(**self.defaults)
        )

    def run(self, module, options):
        """
        Run the operator.

        :param module: The target module path.
        :type  module: ``str``

        :param options: Any runtime options.
        :type  options: ``dict``

        :return: The operator results.
        :rtype: ``dict``
        """
        logger.debug("Running halstead harvester")
        results = {}
        for filename, details in dict(self.harvester.results).items():
            results[filename] = {}
            for instance in details:
                if isinstance(instance, list):
                    for item in instance:
                        function, report = item
                        results[filename][function] = self._report_to_dict(report)
                else:
                    results[filename] = self._report_to_dict(instance)
        return results

    def _report_to_dict(self, report):
        return {
            "h1": report.h1,
            "h2": report.h2,
            "N1": report.N1,
            "N2": report.N2,
            "vocabulary": report.vocabulary,
            "volume": report.volume,
            "length": report.length,
            "effort": report.effort,
            "difficulty": report.difficulty,
        }



# -*- coding: UTF-8 -*-
"""Main command line."""

import click

from pathlib import Path

from wily import logger, __version__
from wily.archivers import resolve_archiver
from wily.cache import exists, get_default_metrics
from wily.config import DEFAULT_CONFIG_PATH, DEFAULT_GRID_STYLE
from wily.config import load as load_config
from wily.decorators import add_version
from wily.helper.custom_enums import ReportFormat
from wily.operators import resolve_operators


@click.group()
@click.version_option(
    __version__, "-V", "--version", message="\U0001F98A %(prog)s, version %(version)s"
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Print debug information, used for development",
)
@click.option(
    "--config",
    default=DEFAULT_CONFIG_PATH,
    help="Path to configuration file, defaults to wily.cfg",
)
@click.option(
    "-p",
    "--path",
    type=click.Path(resolve_path=True),
    default=".",
    help="Root path to the project folder to scan",
)
@click.option(
    "-c",
    "--cache",
    type=click.Path(resolve_path=True),
    help="Override the default cache path (defaults to $HOME/.wily/HASH)",
)
@click.pass_context
@add_version
def cli(ctx, debug, config, path, cache):
    """
    \U0001F98A Inspect and search through the complexity of your source code.

    To get started, run setup:

      $ wily setup

    To reindex any changes in your source code:

      $ wily build <src>

    Then explore basic metrics with:

      $ wily report <file>

    You can also graph specific metrics in a browser with:

      $ wily graph <file> <metric>
    """
    ctx.ensure_object(dict)
    ctx.obj["DEBUG"] = debug
    if debug:
        logger.setLevel("DEBUG")
    else:
        logger.setLevel("INFO")

    ctx.obj["CONFIG"] = load_config(config)
    if path:
        logger.debug(f"Fixing path to {path}")
        ctx.obj["CONFIG"].path = path
    if cache:
        logger.debug(f"Fixing cache to {cache}")
        ctx.obj["CONFIG"].cache_path = cache
    logger.debug(f"Loaded configuration from {config}")


@cli.command()
@click.option(
    "-n",
    "--max-revisions",
    default=None,
    type=click.INT,
    help="The maximum number of historical commits to archive",
)
@click.argument("targets", type=click.Path(resolve_path=True), nargs=-1, required=False)
@click.option(
    "-o",
    "--operators",
    type=click.STRING,
    help="List of operators, separated by commas",
)
@click.option(
    "-a",
    "--archiver",
    type=click.STRING,
    default="git",
    help="Archiver to use, defaults to git if git repo, else filesystem",
)
@click.pass_context
def build(ctx, max_revisions, targets, operators, archiver):
    """Build the wily cache."""
    config = ctx.obj["CONFIG"]

    from wily.commands.build import build

    if max_revisions:
        logger.debug(f"Fixing revisions to {max_revisions}")
        config.max_revisions = max_revisions

    if operators:
        logger.debug(f"Fixing operators to {operators}")
        config.operators = operators.strip().split(",")

    if archiver:
        logger.debug(f"Fixing archiver to {archiver}")
        config.archiver = archiver

    if targets:
        logger.debug(f"Fixing targets to {targets}")
        config.targets = targets

    build(
        config=config,
        archiver=resolve_archiver(config.archiver),
        operators=resolve_operators(config.operators),
    )
    logger.info(
        "Completed building wily history, run `wily report <file>` or `wily index` to see more."
    )


@cli.command()
@click.pass_context
@click.option("--message/--no-message", default=False, help="Include revision message")
def index(ctx, message):
    """Show the history archive in the .wily/ folder."""
    config = ctx.obj["CONFIG"]

    if not exists(config):
        handle_no_cache(ctx)

    from wily.commands.index import index

    index(config=config, include_message=message)


@cli.command()
@click.argument("file", type=click.Path(resolve_path=False))
@click.argument("metrics", nargs=-1, required=False)
@click.option("-n", "--number", help="Number of items to show", type=click.INT)
@click.option("--message/--no-message", default=False, help="Include revision message")
@click.option(
    "-f",
    "--format",
    default=ReportFormat.CONSOLE.name,
    help="Specify report format (console or html)",
    type=click.Choice(ReportFormat.get_all()),
)
@click.option(
    "--console-format",
    default=DEFAULT_GRID_STYLE,
    help="Style for the console grid, see Tabulate Documentation for a list of styles."
)
@click.option(
    "-o", "--output", help="Output report to specified HTML path, e.g. reports/out.html"
)
@click.pass_context
def report(ctx, file, metrics, number, message, format, console_format, output):
    """Show metrics for a given file."""
    config = ctx.obj["CONFIG"]

    if not exists(config):
        handle_no_cache(ctx)

    if not metrics:
        metrics = get_default_metrics(config)
        logger.info(f"Using default metrics {metrics}")

    new_output = Path().cwd()
    if output:
        new_output = new_output / Path(output)
    else:
        new_output = new_output / "wily_report" / "index.html"

    from wily.commands.report import report

    logger.debug(f"Running report on {file} for metric {metrics}")
    logger.debug(f"Output format is {format}")

    report(
        config=config,
        path=file,
        metrics=metrics,
        n=number,
        output=new_output,
        include_message=message,
        format=ReportFormat[format],
        console_format=console_format,
    )


@cli.command()
@click.argument("files", type=click.Path(resolve_path=False), nargs=-1, required=True)
@click.option(
    "--metrics",
    default=None,
    help="comma-seperated list of metrics, see list-metrics for choices",
)
@click.option(
    "--all/--changes-only",
    default=False,
    help="Show all files, instead of changes only",
)
@click.option(
    "--detail/--no-detail",
    default=True,
    help="Show function/class level metrics where available",
)
@click.pass_context
def diff(ctx, files, metrics, all, detail):
    """Show the differences in metrics for each file."""
    config = ctx.obj["CONFIG"]

    if not exists(config):
        handle_no_cache(ctx)

    if not metrics:
        metrics = get_default_metrics(config)
        logger.info(f"Using default metrics {metrics}")
    else:
        metrics = metrics.split(",")
        logger.info(f"Using specified metrics {metrics}")

    from wily.commands.diff import diff

    logger.debug(f"Running diff on {files} for metric {metrics}")
    diff(
        config=config, files=files, metrics=metrics, changes_only=not all, detail=detail
    )


@cli.command()
@click.argument("path", type=click.Path(resolve_path=False))
@click.argument("metrics", nargs=-2, required=True)
@click.option(
    "-o", "--output", help="Output report to specified HTML path, e.g. reports/out.html"
)
@click.option("-x", "--x-axis", help="Metric to use on x-axis, defaults to history.")
@click.option(
    "-a/-c", "--changes/--all", default=True, help="All commits or changes only"
)
@click.pass_context
def graph(ctx, path, metrics, output, x_axis, changes):
    """
    Graph a specific metric for a given file, if a path is given, all files within path will be graphed.

    Some common examples:

    Graph all .py files within src/ for the raw.loc metric

        $ wily graph src/ raw.loc

    Graph test.py against raw.loc and cyclomatic.complexity metrics

        $ wily graph src/test.py raw.loc cyclomatic.complexity

    Graph test.py against raw.loc and raw.sloc on the x-axis

        $ wily graph src/test.py raw.loc --x-axis raw.sloc
    """
    config = ctx.obj["CONFIG"]

    if not exists(config):
        handle_no_cache(ctx)

    from wily.commands.graph import graph

    logger.debug(f"Running report on {path} for metrics {metrics}")
    graph(
        config=config,
        path=path,
        metrics=metrics,
        output=output,
        x_axis=x_axis,
        changes=changes,
    )


@cli.command()
@click.option("-y/-p", "--yes/--prompt", default=False, help="Skip prompt")
@click.pass_context
def clean(ctx, yes):
    """Clear the .wily/ folder."""
    config = ctx.obj["CONFIG"]

    if not exists(config):
        handle_no_cache(ctx)

    if not yes:
        p = input("Are you sure you want to delete wily cache? [y/N]")
        if p.lower() != "y":
            exit(0)

    from wily.cache import clean

    clean(config)


@cli.command("list-metrics")
@click.pass_context
def list_metrics(ctx):
    """List the available metrics."""
    config = ctx.obj["CONFIG"]

    if not exists(config):
        handle_no_cache(ctx)

    from wily.commands.list_metrics import list_metrics

    list_metrics()


@cli.command("setup")
@click.pass_context
def setup(ctx):
    """Run a guided setup to build the wily cache."""
    handle_no_cache(ctx)


def handle_no_cache(context):
    """Handle lack-of-cache error, prompt user for index process."""
    logger.error(
        f"Could not locate wily cache, the cache is required to provide insights."
    )
    p = input("Do you want to run setup and index your project now? [y/N]")
    if p.lower() != "y":
        exit(1)
    else:
        revisions = input("How many previous git revisions do you want to index? : ")
        revisions = int(revisions)
        path = input("Path to your source files; comma-separated for multiple: ")
        paths = path.split(",")
        context.invoke(
            build,
            max_revisions=revisions,
            targets=paths,
            operators=None,
        )


if __name__ == "__main__":
    cli()  # pragma: no cover



import pathlib
from textwrap import dedent
import os
import shutil
import tempfile
import pytest
from click.testing import CliRunner
from git import Repo, Actor

import wily.__main__ as main


@pytest.fixture
def gitdir(tmpdir):
    """ Create a project and add code to it """
    repo = Repo.init(path=tmpdir)
    tmppath = pathlib.Path(tmpdir)
    testpath = tmppath / "src" / "test.py"
    (tmppath / "src").mkdir()
    # Write a test file to the repo
    with open(testpath, "w") as test_txt:
        test_txt.write("import abc")

    index = repo.index
    index.add([str(testpath)])

    author = Actor("An author", "author@example.com")
    committer = Actor("A committer", "committer@example.com")

    index.commit("basic test", author=author, committer=committer)

    first_test = """
    import abc
    foo = 1
    def function1():
        a = 1 + 1
    
    class Class1(object):
        def method(self):
            b = 1 + 5
    """
    with open(testpath, "w") as test_txt:
        test_txt.write(dedent(first_test))

    index.add([str(testpath)])
    index.commit("add line", author=author, committer=committer)

    second_test = """
    import abc
    foo = 1
    def function1():
        a = 1 + 1
    class Class1(object):
        def method(self):
            b = 1 + 5
            if b == 6:
                return 'banana'
    """

    with open(testpath, "w") as test_txt:
        test_txt.write(dedent(second_test))

    index.add([str(testpath)])
    index.commit("remove line", author=author, committer=committer)

    yield tmpdir
    repo.close()


@pytest.fixture
def builddir(gitdir):
    """
    A directory with a wily cache
    """
    tmppath = pathlib.Path(gitdir)
    runner = CliRunner()
    result1 = runner.invoke(
        main.cli, ["--debug", "--path", gitdir, "build", str(tmppath / "src")]
    )
    assert result1.exit_code == 0, result1.stdout

    result2 = runner.invoke(main.cli, ["--debug", "--path", gitdir, "index"])
    assert result2.exit_code == 0, result2.stdout

    return gitdir


@pytest.fixture(autouse=True)
def cache_path(monkeypatch):
    """
    Configure wily cache and home path, clean up cache afterward
    """
    tmp = tempfile.mkdtemp()
    monkeypatch.setenv("HOME", tmp)
    yield tmp
    shutil.rmtree(tmp)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['wily', 'wily.archivers', 'wily.commands', 'wily.helper', 'wily.operators']

package_data = \
{'': ['*'], 'wily': ['templates/*', 'templates/css/*']}

install_requires = \
['gitpython', 'radon', 'click', 'colorlog', 'tabulate', 'plotly', 'progress']

extras_require = \
{":python_version == '3.6'": ['dataclasses']}

entry_points = \
{'console_scripts': ['wily = wily.__main__:cli']}

setup(name='wily',
      version='1.12.1',
      description='Wily.',
      author='Anthony Shaw',
      author_email='anthonyshaw@apache.org',
      url='https://github.com/tonybaloney/wily',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.6',
     )



import json
import pathlib
import sys
import pytest

import wily.cache as cache
from wily.archivers import Revision
from wily.config import DEFAULT_CONFIG, ARCHIVER_GIT


def test_exists(tmpdir):
    """
    Test that exists() returns true if path does exist
    """
    config = DEFAULT_CONFIG
    config.cache_path = tmpdir
    assert cache.exists(config)


def test_exists_older(tmpdir):
    """
    Test that exists() returns true if path does exist
    and has a file with the version older than the current
    """
    config = DEFAULT_CONFIG
    tmp_path = pathlib.Path(tmpdir)
    config.cache_path = tmp_path / ".wily"
    (tmp_path / ".wily").mkdir()
    with open((tmp_path / ".wily" / "index.json"), "w+") as index:
        index.write('{"version": "0.1.0"}')
    assert cache.exists(config)


def test_not_exists():
    """
    Test that exists() returns false if path does not exist
    """
    config = DEFAULT_CONFIG
    config.cache_path = "/v/v/w"
    assert not cache.exists(config)


def test_get_default_metrics_empty(tmpdir):
    """ 
    Test that get_metrics goes ok with an empty index
    """
    config = DEFAULT_CONFIG
    tmppath = pathlib.Path(tmpdir) / ".wily"
    config.cache_path = str(tmppath)
    tmppath.mkdir()
    (tmppath / "git").mkdir()
    with open(tmppath / "git" / "index.json", "w+") as f:
        f.write("[]")

    metrics = cache.get_default_metrics(config)
    assert metrics == []


def test_create_and_delete(tmpdir):
    """
    Test that create() will create a folder with the correct path
    and then delete it.
    """
    config = DEFAULT_CONFIG
    cache_path = pathlib.Path(tmpdir) / ".wily"
    config.cache_path = str(cache_path)
    assert not cache.exists(config)
    cache.create(config)
    assert cache.exists(config)
    cache.clean(config)
    assert not cache.exists(config)


def test_create_when_exists(tmpdir):
    """
    Test that create() will continue if the folder already exists
    """
    config = DEFAULT_CONFIG
    cache_path = pathlib.Path(tmpdir) / ".wily"
    pathlib.Path(cache_path).mkdir()
    config.cache_path = str(cache_path)
    assert cache.exists(config)
    assert str(cache.create(config)) == str(cache_path)


def test_clean_when_not_exists(tmpdir):
    """
    Test that clean() will continue if the folder does not exist
    """
    config = DEFAULT_CONFIG
    cache_path = pathlib.Path(tmpdir) / ".wily"
    config.cache_path = str(cache_path)
    assert not cache.exists(config)
    assert cache.clean(config) is None


def test_store_basic(tmpdir):
    config = DEFAULT_CONFIG
    cache_path = pathlib.Path(tmpdir) / ".wily"
    cache_path.mkdir()
    config.cache_path = cache_path
    target_path = str(pathlib.Path(tmpdir) / "foo" / "bar.py")
    _TEST_STATS = {"operator_data": {"test": {target_path: {"metric1": 1}}}}
    _TEST_REVISION = Revision(
        key="12345",
        author_name="Anthony Shaw",
        author_email="anthony@test.com",
        date="17/01/1990",
        message="my changes",
    )
    fn = cache.store(config, ARCHIVER_GIT, _TEST_REVISION, _TEST_STATS)
    with open(fn) as cache_item:
        result = json.load(cache_item)
        assert isinstance(result, dict)
        assert result == _TEST_STATS


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
def test_store_twice(tmpdir):
    """ Test that you can't write the same revision twice """
    config = DEFAULT_CONFIG
    cache_path = pathlib.Path(tmpdir) / ".wily"
    cache_path.mkdir()
    config.cache_path = cache_path
    target_path = str(pathlib.Path(tmpdir) / "foo" / "bar.py")
    _TEST_STATS = {"operator_data": {"test": {target_path: {"metric1": 1}}}}
    _TEST_REVISION = Revision(
        key="12345",
        author_name="Anthony Shaw",
        author_email="anthony@test.com",
        date="17/01/1990",
        message="my changes",
    )
    fn = cache.store(config, ARCHIVER_GIT, _TEST_REVISION, _TEST_STATS)
    with pytest.raises(RuntimeError):
        cache.store(config, ARCHIVER_GIT, _TEST_REVISION, _TEST_STATS)


def test_store_relative_paths(tmpdir):
    """
    Test that the store command works when absolute paths are used for the targets..
    """
    config = DEFAULT_CONFIG
    cache_path = pathlib.Path(tmpdir) / ".wily"
    target_path = str(pathlib.Path(tmpdir) / "foo" / "bar.py")
    cache_path.mkdir()
    config.cache_path = cache_path
    config.path = tmpdir
    _TEST_STATS = {"operator_data": {"test": {target_path: {"metric1": 1}}}}
    _TEST_REVISION = Revision(
        key="12345",
        author_name="Anthony Shaw",
        author_email="anthony@test.com",
        date="17/01/1990",
        message="my changes",
    )
    fn = cache.store(config, ARCHIVER_GIT, _TEST_REVISION, _TEST_STATS)
    with open(fn) as cache_item:
        result = json.load(cache_item)
        assert isinstance(result, dict)
        if sys.platform == "win32":
            assert "foo\\bar.py" in result["operator_data"]["test"].keys()
        else:
            assert "foo/bar.py" in result["operator_data"]["test"].keys()


def test_store_index(tmpdir):
    """
    Test the store index
    """
    config = DEFAULT_CONFIG
    cache_path = pathlib.Path(tmpdir) / ".wily"
    target_path = pathlib.Path(tmpdir) / "foo" / "bar.py"
    cache_path.mkdir()
    config.cache_path = cache_path
    config.path = tmpdir
    _TEST_INDEX = [{"message": "a", "date": 1234}, {"message": "b", "date": 1345}]
    fn = cache.store_archiver_index(config, ARCHIVER_GIT, _TEST_INDEX)
    with open(fn) as cache_item:
        result = json.load(cache_item)
        assert isinstance(result, list)
        assert result[0] == _TEST_INDEX[1]



import pytest

import wily.operators


def test_resolve_operator():
    op = wily.operators.resolve_operator("cyclomatic")
    assert op == wily.operators.OPERATOR_CYCLOMATIC


def test_resolve_bad_operator():
    with pytest.raises(ValueError):
        wily.operators.resolve_operator("banana")


def test_resolve_operators():
    ops = wily.operators.resolve_operators(("cyclomatic", "raw"))
    assert ops[0] == wily.operators.OPERATOR_CYCLOMATIC
    assert ops[1] == wily.operators.OPERATOR_RAW


def test_resolve_metric():
    metric = wily.operators.resolve_metric("raw.loc")
    assert metric.name == "loc"


def test_resolve_invalid_metric():
    with pytest.raises(ValueError):
        wily.operators.resolve_metric("raw.spanner")


def test_resolve_short_metric():
    metric = wily.operators.resolve_metric("loc")
    assert metric.name == "loc"



"""
Parameterised tests for each operator (and some combinations).

Build them and test out some of the metrics/commands work correctly.
"""
import pytest
import sys
from click.testing import CliRunner
import pathlib
from textwrap import dedent

import wily.__main__ as main

_path = "src\\test.py" if sys.platform == "win32" else "src/test.py"


operators = (
    "halstead",
    "cyclomatic",
    "maintainability",
    "raw",
    "halstead,cyclomatic",
    "maintainability,raw",
    "halstead,cyclomatic,maintainability,raw",
)


@pytest.mark.parametrize("operator", operators)
def test_operator(operator, gitdir):
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--debug", "--path", gitdir, "build", "src", "-o", operator]
    )
    assert result.exit_code == 0, result.stdout

    result = runner.invoke(main.cli, ["--debug", "--path", gitdir, "report", _path])
    assert result.exit_code == 0, result.stdout

    result = runner.invoke(
        main.cli, ["--debug", "--path", gitdir, "diff", _path, "--all"]
    )
    assert result.exit_code == 0, result.stdout
    assert "test.py" in result.stdout

    complex_test = """
            import abc
            foo = 1
            def function1():
                a = 1 + 1
                if a == 2:
                    print(1)
            class Class1(object):
                def method(self):
                    b = 1 + 5
                    if b == 6:
                        if 1==2:
                           if 2==3:
                              print(1)
            """

    with open(pathlib.Path(gitdir) / "src" / "test.py", "w") as test_py:
        test_py.write(dedent(complex_test))

    result = runner.invoke(
        main.cli, ["--debug", "--path", gitdir, "diff", _path, "--all"]
    )
    assert result.exit_code == 0, result.stdout
    assert "test.py" in result.stdout



import pathlib

from click.testing import CliRunner

import wily.__main__ as main


def test_list_metrics(builddir):
    """
    Test that list-metrics works with a build
    """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", builddir, "list-metrics"])
    assert result.exit_code == 0, result.stdout


def test_clean(builddir):
    """ Test the clean feature """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", builddir, "clean", "--yes"])
    assert result.exit_code == 0, result.stdout
    cache_path = pathlib.Path(builddir) / ".wily"
    assert not cache_path.exists()


def test_clean_no_cache(tmpdir):
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", tmpdir, "clean"])
    assert result.exit_code == 1, result.stdout


def test_list_metrics_no_cache(tmpdir):
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", tmpdir, "list-metrics"])
    assert result.exit_code == 1, result.stdout



"""
Filesystem Archiver.

Implementation of the archiver API for a standard directory (no revisions)
"""
import logging
import os.path
import hashlib
from wily.archivers import BaseArchiver, Revision

logger = logging.getLogger(__name__)


class FilesystemArchiver(BaseArchiver):
    """Basic implementation of the base archiver."""

    name = "filesystem"

    def __init__(self, config):
        """
        Instantiate a new Filesystem Archiver.

        :param config: The wily configuration
        :type  config: :class:`wily.config.WilyConfig`
        """
        self.config = config

    def revisions(self, path, max_revisions):
        """
        Get the list of revisions.

        :param path: the path to target.
        :type  path: ``str``

        :param max_revisions: the maximum number of revisions.
        :type  max_revisions: ``int``

        :return: A list of revisions.
        :rtype: ``list`` of :class:`Revision`
        """
        mtime = os.path.getmtime(path)
        key = hashlib.sha1(str(mtime).encode()).hexdigest()[:7]
        return [
            Revision(
                key=key,
                author_name="Local User",  # Don't want to leak local data
                author_email="-",  # as above
                date=int(mtime),
                message="None",
            )
        ]

    def checkout(self, revision, options):
        """
        Checkout a specific revision.

        :param revision: The revision identifier.
        :type  revision: :class:`Revision`

        :param options: Any additional options.
        :type  options: ``dict``
        """
        # effectively noop since there are no revision
        pass



"""A module containing custom enums for wily."""
from enum import Enum


class ReportFormat(Enum):
    """Represent the available report formats."""

    CONSOLE = 1
    HTML = 2

    @classmethod
    def get_all(cls):
        """Return a list with all Enumerations."""
        return [format.name for format in cls]
